# -*- coding: utf-8; -*-
"""add product.average_weight

Revision ID: c8981df473d0
Revises: efb7cd318947
Create Date: 2020-09-02 13:13:42.629959

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c8981df473d0'
down_revision = u'efb7cd318947'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.add_column('product', sa.Column('average_weight', sa.Numeric(precision=12, scale=3), nullable=True))
    op.add_column('product_version', sa.Column('average_weight', sa.Numeric(precision=12, scale=3), autoincrement=False, nullable=True))


def downgrade():

    # product
    op.drop_column('product_version', 'average_weight')
    op.drop_column('product', 'average_weight')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
TrainWreck -> "self" data importing
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail import importing
from rattail.trainwreck.importing.trainwreck import FromTrainwreckHandler
from rattail.trainwreck import importing as trainwreck_importing


class FromTrainwreckToSelf(FromTrainwreckHandler, importing.ToSQLAlchemyHandler):
    """
    Base handler class for Trainwreck -> "self"
    """
    host_key = 'self'
    host_title = "Trainwreck (self)"
    generic_host_title = "Trainwreck (self)"

    local_key = 'trainwreck'
    local_title = "Trainwreck (self)"
    generic_local_title = "Trainwreck (self)"

    def begin_local_transaction(self):
        self.session = self.host_session

    def rollback_transaction(self):
        self.rollback_host_transaction()

    def commit_transaction(self):
        self.commit_host_transaction()


class FromTrainwreck(importing.FromSQLAlchemy):
    """
    Common base class for the "host" side of importers which read data
    from the Trainwreck DB for the sake of updating misc. other tables
    in that same DB.
    """


class TransactionImporter(FromTrainwreck, trainwreck_importing.model.TransactionImporter):
    """
    Base class for Trainwreck -> "self" for Transaction model.
    """
    allow_create = False
    allow_delete = False

    @property
    def host_model_class(self):
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.Transaction

    def query(self):
        query = super(TransactionImporter, self).query()
        trainwreck = self.config.get_trainwreck_model()

        time_field = getattr(trainwreck.Transaction, self.match_on_time_field)
        query = query.filter(time_field >= self.app.make_utc(self.start_time))\
                     .filter(time_field < self.app.make_utc(self.end_time))

        query = query.options(orm.joinedload(trainwreck.Transaction.items)\
                              .joinedload(trainwreck.TransactionItem.discounts))

        return query

    def normalize_host_object(self, txn):
        """
        This method is defined only as a convenience for the simple
        case where you need to mark a transaction as having been
        updated, along with some other "processing" logic.
        """
        return {
            'uuid': txn.uuid,
            'self_updated': True,
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Importers for SQLAlchemy
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm
from sqlalchemy.orm.exc import NoResultFound

from rattail.importing import Importer, FromQuery


class FromSQLAlchemy(FromQuery):
    """
    Base class for importers whose external data source is a SQLAlchemy query.
    """
    host_session = None

    # For default behavior, set this to a model class to be used in generating
    # the host data query.
    host_model_class = None

    def query(self):
        """
        Must return the primary query which will define the data set.  Default
        behavior is to leverage :attr:`host_session` and generate a query for
        the class defined by :attr:`host_model_class`.
        """
        return self.host_session.query(self.host_model_class)


class ToSQLAlchemy(Importer):
    """
    Base class for all data importers which support the common use case of
    targeting a SQLAlchemy ORM on the local side.  This is the base class for
    all primary Rattail importers.
    """
    caches_local_data = True

    @property
    def model_mapper(self):
        """
        Reference to the effective SQLAlchemy mapper for the local model class.
        """
        return orm.class_mapper(self.model_class)

    @property
    def model_table(self):
        """
        Reference to the effective SQLAlchemy table for the local model class.
        """
        tables = self.model_mapper.tables
        assert len(tables) == 1
        return tables[0]

    @property
    def simple_fields(self):
        """
        Returns the list of column names on the underlying local model mapper.
        """
        return list(self.model_mapper.columns.keys())

    @property
    def supported_fields(self):
        """
        All/only simple fields are supported by default.
        """
        return self.simple_fields

    def get_single_local_object(self, key):
        """
        Try to fetch the object from the local database, using SA ORM.
        """
        query = self.session.query(self.model_class)
        for i, k in enumerate(self.key):
            query = query.filter(getattr(self.model_class, k) == key[i])
        try:
            return query.one()
        except NoResultFound:
            pass

    def create_object(self, key, host_data):
        """
        Create and return a new local object for the given key, fully populated
        from the given host data.  This may return ``None`` if no object is
        created.

        Note that this also adds the new object to the local database session.
        """
        with self.session.no_autoflush:
            obj = super(ToSQLAlchemy, self).create_object(key, host_data)
        if obj:
            self.session.add(obj)
            return obj

    def delete_object(self, obj):
        """
        Delete the given object from the local system (or not), and return a
        boolean indicating whether deletion was successful.  Default logic will
        truly delete and expunge the local object from the session.
        """
        self.session.delete(obj)
        return True

    def cache_local_data(self, host_data=None):
        """
        Cache all local objects and data using SA ORM.
        """
        return self.cache_model(self.model_class, key=self.get_cache_key,
                                query=self.cache_query(),
                                # omit_duplicates=True,
                                query_options=self.cache_query_options(),
                                normalizer=self.normalize_cache_object,
                                message=self.cache_local_message())

    def cache_query(self):
        """
        Return the query to be used when caching "local" data.
        """
        return self.session.query(self.model_class)

    def cache_query_options(self):
        """
        Return a list of options to apply to the cache query, if needed.
        """
        return []

    def flush_create_update(self):
        """
        Flush the database session, to send SQL to the server for all changes
        made thus far.
        """
        self.session.flush()


class FromSQLAlchemySameToSame(FromSQLAlchemy):
    """
    Special base class for importers which sync data from one database to
    another, when the schema for each database is identical.
    """

    @property
    def host_model_class(self):
        return self.model_class

    @property
    def supported_fields(self):
        """
        We only need to support the simple fields, since all relevant tables
        should be covered and therefore no need to do crazy foreign key
        acrobatics etc.
        """
        return self.simple_fields

    def query(self):
        """
        Leverage the same caching optimizations on both sides, if applicable.
        """
        query = super(FromSQLAlchemySameToSame, self).query()

        if hasattr(self, 'cache_query_options'):
            options = self.cache_query_options()
            if options:
                for option in options:
                    query = query.options(option)

        return query

    def normalize_host_object(self, obj):
        """
        Normalization should work the same for both sides.
        """
        return self.normalize_local_object(obj)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Report Handlers
"""

from __future__ import unicode_literals, absolute_import

import datetime
import decimal

import six

from rattail.db import model
from rattail.util import load_entry_points, load_object
from rattail.mail import send_email
from rattail.time import localtime


class ReportHandler(object):
    """
    Base class for all report handlers.  Also provides default implementation,
    which is assumed to be sufficient for most needs.
    """
    entry_point_section = 'rattail.reports'

    def __init__(self, config=None, **kwargs):
        self.config = config
        self.app = config.get_app()
        self.enum = config.get_enum() if config else None
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_reports(self):
        """
        Returns a dict of all available reports, keyed by the report
        type key.

        By default this includes all reports which are properly
        registered via setuptools entry points.  It also will include
        any custom Poser reports, although that can be disabled via
        config if desired.
        """
        # properly registered reports
        reports = load_entry_points(self.entry_point_section)

        # maybe also include poser reports
        if self.config.getbool('rattail.reporting',
                               'include_poser_reports',
                               default=True):

            # these come back in different format so must normalize
            poser_handler = self.app.get_poser_handler()
            for report_info in poser_handler.get_all_reports():
                if not report_info.get('error'):
                    report = report_info['report']
                    reports[report.type_key] = report

        return reports

    def get_report(self, key):
        """
        Fetch a report by key.  If the report can be located, this will return an
        instance thereof; otherwise returns ``None``.
        """
        report = self.get_reports().get(key)
        if report:
            return report(self.config)

    def generate_output(self, session, report, params, user, progress=None, **kwargs):
        """
        Generate and return output for the given report and params.
        """
        data = report.make_data(session, params, progress=progress, **kwargs)

        output = model.ReportOutput()
        output.id = self.app.next_counter_value(session, 'batch_id')
        output.report_name = report.make_report_name(session, params, data, **kwargs)
        output.report_type = report.type_key
        output.params = self.safe_params(**params)
        output.filename = report.make_filename(session, params, data, **kwargs)
        output.created_by = user
        session.add(output)
        session.flush()

        # nb. stash data for the report
        # TODO: this is an undocumented feature...
        output._generated_data = data

        path = output.filepath(self.config, makedirs=True)
        report.write_file(session, params, data, path, progress=progress, **kwargs)
        return output

    def safe_params(self, **kwargs):
        params = {}
        for key, value in six.iteritems(kwargs):
            if isinstance(value, datetime.date):
                value = six.text_type(value)
            elif isinstance(value, decimal.Decimal):
                value = float(value)
            params[key] = value
        return params

    def email_output(self, report, output, mailkey,
                     fallback_mailkey='common_report', extra_data={}, **kwargs):
        """
        Send an email (using the given :paramref:`mailkey`) with the report
        output as an attachment.

        :param report: The relevant :class:`Report` instance.
        :param output: The relevant :class:`ReportOutput` instance.
        :param mailkey: Config key which identifies the type of email to be
           sent.
        :param fallback_mailkey: Config key to be used as fallback, should the
           :paramref:`mailkey` not point to a valid email config.
        :param extra_data: Additional context data to be passed along to the
           email template.
        :param kwargs: Additional kwargs to be passed directly to
           :func:`~rattail.mail.send_email()`.
        """
        if 'attachments' not in kwargs:
            path = output.filepath(self.config)
            kwargs['attachments'] = [path]

        data = {
            'config': self.config,
            'handler': self,
            'report': report,
            'output': output,
            'localtime': localtime,
        }

        if extra_data:
            data.update(extra_data)

        if 'params' not in data:
            data['params'] = output.params

        send_email(self.config, mailkey, data, fallback_key=fallback_mailkey,
                   **kwargs)


def get_report_handler(config, **kwargs):
    """
    Create and return the configured :class:`ReportHandler` instance.
    """
    spec = config.get('rattail.reports', 'handler')
    if spec:
        return load_object(spec)(config)
    return ReportHandler(config)



# -*- coding: utf-8 -*-
"""add purchase.department

Revision ID: 4bd528bca236
Revises: 9740f0dcf84b
Create Date: 2016-12-09 12:21:32.663614

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '4bd528bca236'
down_revision = u'9740f0dcf84b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase
    op.add_column('purchase', sa.Column('department_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'purchase_fk_department', 'purchase', 'department', ['department_uuid'], ['uuid'])

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('department_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'purchase_batch_fk_department', 'purchase_batch', 'department', ['department_uuid'], ['uuid'])


def downgrade():

    # purchase_batch
    op.drop_constraint(u'purchase_batch_fk_department', 'purchase_batch', type_='foreignkey')
    op.drop_column('purchase_batch', 'department_uuid')

    # purchase_batch
    op.drop_constraint(u'purchase_fk_department', 'purchase', type_='foreignkey')
    op.drop_column('purchase', 'department_uuid')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Commands
"""

import importlib
import os
import sys
import json
import platform
import argparse
import datetime
import socket
import shutil
import subprocess
import warnings
import logging
from getpass import getpass

import humanize

from rattail import __version__
from rattail.core import Object
from rattail.util import load_entry_points, load_object
from rattail.progress import ConsoleProgress, SocketProgress
from rattail.config import make_config, parse_list
from rattail.util import progress_loop
from rattail.time import make_utc
from rattail.db.config import configure_versioning


log = logging.getLogger(__name__)


class ArgumentParser(argparse.ArgumentParser):
    """
    Custom argument parser.

    This overrides some of the parsing logic which is specific to the primary
    command object.
    """

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        args.argv = argv
        return args


def date_argument(string):
    """
    Validate and coerce a date argument.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--date', type=date_argument)
    """
    try:
        date = datetime.datetime.strptime(string, '%Y-%m-%d').date()
    except ValueError:
        raise argparse.ArgumentTypeError("Date must be in YYYY-MM-DD format")
    return date


def dict_argument(string):
    """
    Coerce the given string to a Python dictionary.  The string is assumed to
    be JSON-encoded; this function merely invokes ``json.loads()`` on it.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--date', type=dict_argument)
    """
    try:
        return json.loads(string)
    except json.decoder.JSONDecodeError:
        raise argparse.ArgumentTypeError("Argument must be valid JSON-encoded string")


def list_argument(string):
    """
    Coerce the given string to a list of strings, splitting on whitespace
    and/or commas.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--things', type=list_argument)
    """
    return parse_list(string)


class Command(Object):
    """
    The primary command for the application.

    This effectively *is* the ``rattail`` console application.  It mostly
    provides the structure for subcommands, which really do all the work.

    This command is designed to be subclassed, should your application need
    similar functionality.
    """
    name = 'rattail'
    version = __version__
    description = "Rattail Software Framework"
    long_description = """
Rattail is a retail software framework.

Copyright (c) 2010-2015 Lance Edgar <lance@edbob.org>

This program comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.
See the file COPYING.txt for more information.
"""

    stdout = sys.stdout
    stderr = sys.stderr

    def __init__(self, **kwargs):
        super(Command, self).__init__(**kwargs)
        self.subcommands = load_entry_points('{}.commands'.format(self.name.replace('-', '_')))

    def __str__(self):
        return self.name

    @property
    def db_config_section(self):
        """
        Name of section in config file which should have database connection
        info.  This defaults to ``'rattail.db'`` but may be overridden so the
        command framework can sit in front of a non-Rattail database if needed.

        This is used to auto-configure a "default" database engine for the app,
        when any command is invoked.
        """
        return 'rattail.db'

    @property
    def db_session_factory(self):
        """
        Reference to the "primary" ``Session`` class, which will be configured
        automatically during app startup.  Defaults to :class:`rattail.db.Session`.
        """
        from rattail.db import Session
        return Session

    @property
    def db_model(self):
        """
        Reference to the Python module which is to be used as the primary data
        model.  Defaults to ``rattail.db.model``.
        """
        return self.config.get_model()

    def iter_subcommands(self):
        """
        Iterate over the subcommands.

        This is a generator which yields each associated :class:`Subcommand`
        class sorted by :attr:`Subcommand.name`.
        """
        for name in sorted(self.subcommands):
            yield self.subcommands[name]

    def print_help(self):
        """
        Print help text for the primary command.

        The output will include a list of available subcommands.
        """
        # TODO: this should leverage parser config...
        self.stdout.write("""{0.description}

Usage: {0.name} [options] <command> [command-options]

Options:
  -c PATH, --config=PATH
                    Config path (may be specified more than once)
  -n, --no-init     Do not read any config files
  -P, --progress    Show progress indicators (where relevant)
  -R, --runas       Optional username to impersonate when running command
  -V, --version     Display program version and exit

Commands:\n""".format(self))

        for cmd in self.iter_subcommands():
            self.stdout.write("  {0:<16s}  {1}\n".format(cmd.name, cmd.description))

        self.stdout.write("\nTry '{0} help <command>' for more help.\n".format(self.name))

    def run(self, *args):
        """
        Parse command line arguments and execute appropriate subcommand.
        """
        parser = ArgumentParser(
            prog=self.name,
            description=self.description,
            add_help=False,
            )

        parser.add_argument('-c', '--config', action='append', dest='config_paths',
                            metavar='PATH')

        # TODO: i think these aren't really being used in practice..?
        parser.add_argument('-n', '--no-init', action='store_true', default=False)
        parser.add_argument('--no-extend-config', dest='extend_config', action='store_false')

        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('-P', '--progress', action='store_true', default=False)
        parser.add_argument('--progress-socket',
                            help="Optional socket (e.g. localhost:8487) to which progress info should be written.")
        parser.add_argument('--runas', '-R', metavar='USERNAME',
                            help="Optional username to impersonate when running the command.  "
                            "This is only relevant for / used by certain commands.")
        parser.add_argument('--stdout', metavar='PATH', type=argparse.FileType('w'),
                            help="Optional path to which STDOUT should be effectively redirected.")
        parser.add_argument('--stderr', metavar='PATH', type=argparse.FileType('w'),
                            help="Optional path to which STDERR should be effectively redirected.")

        # data versioning
        parser.add_argument('--versioning', action='store_true',
                            help="Force *enable* of data versioning.  If set, then --no-versioning "
                            "cannot also be set.  If neither is set, config will determine whether "
                            "or not data versioning should be enabled.")
        parser.add_argument('--no-versioning', action='store_true',
                            help="Force *disable* of data versioning.  If set, then --versioning "
                            "cannot also be set.  If neither is set, config will determine whether "
                            "or not data versioning should be enabled.")

        parser.add_argument('-V', '--version', action='version',
                            version="%(prog)s {0}".format(self.version))
        parser.add_argument('command', nargs='*')

        # Parse args and determine subcommand.
        args = parser.parse_args(list(args))
        if not args or not args.command:
            self.print_help()
            sys.exit(1)

        # TODO: can we make better args so this is handled by argparse somehow?
        if args.versioning and args.no_versioning:
            sys.stderr.write("Cannot pass both --versioning and --no-versioning\n")
            sys.exit(1)

        # Show (sub)command help if so instructed, or unknown subcommand.
        cmd = args.command.pop(0)
        if cmd == 'help':
            if len(args.command) != 1:
                self.print_help()
                return
            cmd = args.command[0]
            if cmd not in self.subcommands:
                self.print_help()
                return
            cmd = self.subcommands[cmd](parent=self)
            cmd.parser.print_help()
            return
        elif cmd in self.subcommands:
            if '-h' in args.argv or '--help' in args.argv:
                cmd = self.subcommands[cmd](parent=self)
                cmd.parser.print_help()
                return
        else:
            self.print_help()
            sys.exit(1)

        # Okay, we should be done needing to print help messages.  Now it's
        # safe to redirect STDOUT/STDERR, if necessary.
        if args.stdout:
            self.stdout = args.stdout
        if args.stderr:
            self.stderr = args.stderr

        # if args say not to "init" then we make a sort of empty config
        if args.no_init:
            self.config = make_config([], extend=False, versioning=False)

        else: # otherwise we make a proper config, and maybe turn on versioning
            logging.basicConfig()
            self.config = make_config(args.config_paths, extend=args.extend_config, versioning=False)
            if args.versioning:
                configure_versioning(self.config, force=True)
            elif not args.no_versioning:
                configure_versioning(self.config)

        # import our primary data model now, just in case it hasn't fully been
        # imported yet.  this it to be sure association proxies and the like
        # are fully wired up in the case of extensions
        # TODO: what about web apps etc.? i guess i was only having the problem
        # for some importers, e.g. basic CORE API -> Rattail w/ the schema
        # extensions in place from rattail-corepos
        try:
            self.config.get_model()
        except ImportError:
            pass

        # instantiate the subcommand
        subcmd = self.subcommands[cmd](self, self.config)

        # figure out if/how subcommand should show progress
        subcmd.show_progress = args.progress
        subcmd.progress = None
        if subcmd.show_progress:
            if args.progress_socket:
                host, port = args.progress_socket.split(':')
                subcmd.progress = SocketProgress(host, int(port))
            else:
                subcmd.progress = ConsoleProgress

        # maybe should be verbose
        subcmd.verbose = args.verbose

        # TODO: make this default to something from config?
        subcmd.runas_username = args.runas or None

        # and finally, run the subcommand
        log.debug("running '%s %s' with args: %s", self.name, subcmd.name, args.argv)
        log.debug("full command line was: %s", sys.argv)
        subcmd._run(*(args.command + args.argv))


class Subcommand(object):
    """
    Base class for application subcommands.
    """
    name = 'UNDEFINED'
    description = 'UNDEFINED'

    def __init__(self, parent=None, config=None, show_progress=None):
        self.parent = parent
        self.config = config
        if self.config:
            self.app = self.config.get_app()
        self.stdout = getattr(parent, 'stdout', sys.stdout)
        self.stderr = getattr(parent, 'stderr', sys.stderr)
        self.show_progress = show_progress
        self.progress = ConsoleProgress if show_progress else None
        self.parser = argparse.ArgumentParser(
            prog='{0} {1}'.format(getattr(self.parent, 'name', 'UNDEFINED'), self.name),
            description=self.description)
        self.add_parser_args(self.parser)

    def __repr__(self):
        return "Subcommand(name={0})".format(repr(self.name))

    @property
    def model(self):
        return self.parent.db_model

    def add_parser_args(self, parser):
        """
        Configure additional arguments for the subcommand argument parser.
        """
        pass

    def make_session(self):
        session = self.parent.db_session_factory()
        user = self.get_runas_user(session=session)
        if user:
            session.set_continuum_user(user)
        return session

    def finalize_session(self, session, dry_run=False, success=True):
        """
        Wrap up the given session, per the given arguments.  This is meant to
        provide a simple convenience, for commands which must do work within a
        DB session, but would like to support a "dry run" mode.
        """
        if dry_run:
            session.rollback()
            log.info("dry run, so transaction was aborted")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

    def get_runas_user(self, session=None, username=None):
        """
        Returns a proper User object, which the app should "run as"
        """
        from sqlalchemy import orm
        from rattail.db.util import short_session

        if username is None:
            if hasattr(self, 'runas_username'):
                username = self.runas_username
            if not username and self.config:
                username = self.config.get('rattail', 'runas.default')
        if username:
            user = None
            with short_session(session) as s:
                try:
                    user = s.query(self.model.User).filter_by(username=username).one()
                except orm.exc.NoResultFound:
                    pass
                else:
                    if not session:
                        s.expunge(user)
            return user

    def progress_loop(self, func, items, factory=None, **kwargs):
        return progress_loop(func, items, factory or self.progress, **kwargs)
            
    def _run(self, *args):
        args = self.parser.parse_args(list(args))
        return self.run(args)

    def run(self, args):
        """
        Run the subcommand logic.
        """
        raise NotImplementedError

    def require_prompt_toolkit(self):
        try:
            import prompt_toolkit
        except ImportError:
            value = input("\nprompt_toolkit is not installed.  shall i install it? [Yn] ")
            value = value.strip()
            if value and not self.config.parse_bool(value):
                self.stderr.write("prompt_toolkit is required; aborting\n")
                sys.exit(1)

            subprocess.check_call(['pip', 'install', 'prompt_toolkit'])

    def require_rich(self):
        try:
            import rich
        except ImportError:
            value = input("\nrich is not installed.  shall i install it? [Yn] ")
            value = value.strip()
            if value and not self.config.parse_bool(value):
                self.stderr.write("rich is required; aborting\n")
                sys.exit(1)

            subprocess.check_call(['pip', 'install', 'rich'])

    def rprint(self, *args, **kwargs):
        self.require_rich()

        # TODO: this could look different once python2 is out of the
        # picture; but must avoid `print` keyword for python2
        import rich
        rprint = getattr(rich, 'print')
        return rprint(*args, **kwargs)

    def basic_prompt(self, info, default=None, is_password=False, is_bool=False,
                     required=False):
        self.require_prompt_toolkit()

        from prompt_toolkit import prompt
        from prompt_toolkit.styles import Style

        # message formatting styles
        style = Style.from_dict({
            '': '',
            'bold': 'bold',
        })

        # build prompt message
        message = [
            ('', '\n'),
            ('class:bold', info),
        ]
        if default is not None:
            if is_bool:
                message.append(('', ' [{}]: '.format('Y' if default else 'N')))
            else:
                message.append(('', ' [{}]: '.format(default)))
        else:
            message.append(('', ': '))

        # prompt user for input
        try:
            text = prompt(message, style=style, is_password=is_password)
        except (KeyboardInterrupt, EOFError):
            self.rprint("\n\t[bold yellow]operation canceled by user[/bold yellow]\n",
                        file=self.stderr)
            sys.exit(2)

        if is_bool:
            if text == '':
                return default
            elif text.upper() == 'Y':
                return True
            elif text.upper() == 'N':
                return False
            self.rprint("\n\t[bold yellow]ambiguous, please try again[/bold yellow]\n")
            return self.basic_prompt(info, default, is_bool=True)

        if required and not text and not default:
            return self.basic_prompt(info, default, is_password=is_password,
                                     required=True)

        return text or default


class CheckDatabase(Subcommand):
    """
    Do basic sanity checks on a Rattail DB
    """
    name = 'checkdb'
    description = __doc__.strip()

    def run(self, args):
        import sqlalchemy as sa

        try:
            with self.config.rattail_engine.begin() as cxn:
                cxn.execute(sa.text("select version()"))
        except sa.exc.OperationalError as e:
            self.stderr.write("\nfailed to connect to DB!\n\n{}\n".format(e))
            sys.exit(1)

        self.stdout.write("All checks passed.\n")


class CloneDatabase(Subcommand):
    """
    Clone (supported) data from a source DB to a target DB
    """
    name = 'clonedb'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('source_engine',
                            help="SQLAlchemy engine URL for the source database.")
        parser.add_argument('target_engine',
                            help="SQLAlchemy engine URL for the target database.")
        parser.add_argument('-m', '--model', default='rattail.db.model',
                            help="Dotted path of Python module which contains the data model.")
        parser.add_argument('-C', '--classes', nargs='*',
                            help="Model classes which should be cloned.  Possible values here "
                            "depends on which module contains the data model.  If no classes "
                            "are specified, all available will be cloned.")

    def run(self, args):
        from sqlalchemy import create_engine, orm

        model = importlib.import_module(args.model)
        classes = args.classes
        assert classes

        source_engine = create_engine(args.source_engine)
        target_engine = create_engine(args.target_engine)
        model.Base.metadata.drop_all(bind=target_engine)
        model.Base.metadata.create_all(bind=target_engine)

        Session = orm.sessionmaker()
        src_session = Session(bind=source_engine)
        dst_session = Session(bind=target_engine)

        for clsname in classes:
            log.info("cloning data for model: %s", clsname)
            cls = getattr(model, clsname)
            src_query = src_session.query(cls)
            count = src_query.count()
            log.debug("found %d %s records to clone", count, clsname)
            if not count:
                continue

            mapper = orm.class_mapper(cls)
            key_query = src_session.query(*mapper.primary_key)

            prog = None
            if self.progress:
                prog = self.progress("Cloning data for model: {0}".format(clsname), count)
            for i, key in enumerate(key_query, 1):

                src_instance = src_query.get(key)
                dst_session.merge(src_instance)
                dst_session.flush()

                if prog:
                    prog.update(i)
            if prog:
                prog.destroy()

        src_session.close()
        dst_session.commit()
        dst_session.close()


class EmailBouncer(Subcommand):
    """
    Interacts with the email bouncer daemon.  This command expects a
    subcommand; one of the following:

    * ``rattail bouncer start``
    * ``rattail bouncer stop``
    """
    name = 'bouncer'
    description = "Manage the email bouncer daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/bouncer.pid',
                            help="Path to PID file.")
        parser.add_argument('--daemonize', action='store_true',
                            help="DEPRECATED")
        parser.add_argument('--no-daemonize',
                            '-D', '--do-not-daemonize',
                            action='store_false', dest='daemonize',
                            help="DEPRECATED")

    def run(self, args):
        from rattail.bouncer.daemon import BouncerDaemon

        daemon = BouncerDaemon(args.pidfile, config=self.config)
        if args.subcommand == 'stop':
            daemon.stop()
        else: # start
            try:
                daemon.start(daemonize=False)
            except KeyboardInterrupt:
                self.stderr.write("Interrupted.\n")


class DateOrganize(Subcommand):
    """
    Organize files in a given directory, according to date
    """
    name = 'date-organize'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('folder', metavar='PATH',
                            help="Path to directory containing files which are "
                            "to be organized by date.")

    def run(self, args):
        today = datetime.date.today()
        for filename in sorted(os.listdir(args.folder)):
            path = os.path.join(args.folder, filename)
            if os.path.isfile(path):
                mtime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
                if mtime.date() < today:
                    datedir = mtime.strftime(os.sep.join(('%Y', '%m', '%d')))
                    datedir = os.path.join(args.folder, datedir)
                    if not os.path.exists(datedir):
                        os.makedirs(datedir)
                    shutil.move(path, datedir)


class FileMonitorCommand(Subcommand):
    """
    Interacts with the file monitor service; called as ``rattail filemon``.
    This command expects a subcommand; one of the following:

    * ``rattail filemon start``
    * ``rattail filemon stop``

    On Windows platforms, the following additional subcommands are available:

    * ``rattail filemon install``
    * ``rattail filemon uninstall`` (or ``rattail filemon remove``)

    .. note::
       The Windows Vista family of operating systems requires you to launch
       ``cmd.exe`` as an Administrator in order to have sufficient rights to
       run the above commands.

    .. See :doc:`howto.use_filemon` for more information.
    """

    name = 'filemon'
    description = "Manage the file monitor daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform in ('linux', 'linux2'):
            parser.add_argument('-p', '--pidfile',
                                help="Path to PID file.", metavar='PATH')
            parser.add_argument('--daemonize', action='store_true',
                                help="DEPRECATED")
            parser.add_argument('--no-daemonize',
                                '-D', '--do-not-daemonize',
                                action='store_false', dest='daemonize',
                                help="DEPRECATED")

        elif sys.platform == 'win32': # pragma no cover

            install = subparsers.add_parser('install', help="Install service")
            install.set_defaults(subcommand='install')
            install.add_argument('-a', '--auto-start', action='store_true',
                                 help="Configure service to start automatically.")
            install.add_argument('-U', '--username',
                                 help="User account under which the service should run.")

            remove = subparsers.add_parser('remove', help="Uninstall (remove) service")
            remove.set_defaults(subcommand='remove')

            uninstall = subparsers.add_parser('uninstall', help="Uninstall (remove) service")
            uninstall.set_defaults(subcommand='remove')

    def run(self, args):
        if sys.platform in ('linux', 'linux2'):
            from rattail.filemon import linux as filemon

            if args.subcommand == 'start':
                filemon.start_daemon(self.config, args.pidfile)

            elif args.subcommand == 'stop':
                filemon.stop_daemon(self.config, args.pidfile)

        elif sys.platform == 'win32': # pragma no cover
            self.run_win32(args)

        else:
            self.stderr.write("File monitor is not supported on platform: {0}\n".format(sys.platform))
            sys.exit(1)

    def run_win32(self, args): # pragma no cover
        from rattail.win32 import require_elevation
        from rattail.win32 import service
        from rattail.win32 import users
        from rattail.filemon import win32 as filemon

        require_elevation()

        options = []
        if args.subcommand == 'install':

            username = args.username
            if username:
                if '\\' in username:
                    server, username = username.split('\\')
                else:
                    server = socket.gethostname()
                if not users.user_exists(username, server):
                    sys.stderr.write("User does not exist: {0}\\{1}\n".format(server, username))
                    sys.exit(1)

                password = ''
                while password == '':
                    password = getpass(str("Password for service user: ")).strip()
                options.extend(['--username', r'{0}\{1}'.format(server, username)])
                options.extend(['--password', password])

            if args.auto_start:
                options.extend(['--startup', 'auto'])

        service.execute_service_command(filemon, args.subcommand, *options)

        # If installing with custom user, grant "logon as service" right.
        if args.subcommand == 'install' and args.username:
            users.allow_logon_as_service(username)

        # TODO: Figure out if the following is even required, or if instead we
        # should just be passing '--startup delayed' to begin with?

        # If installing auto-start service on Windows 7, we should update
        # its startup type to be "Automatic (Delayed Start)".
        # TODO: Improve this check to include Vista?
        if args.subcommand == 'install' and args.auto_start:
            if platform.release() == '7':
                name = filemon.RattailFileMonitor._svc_name_
                service.delayed_auto_start_service(name)


class MailMonitorCommand(Subcommand):
    """
    Interacts with the mail monitor service; called as ``rattail
    mailmon``.  This command expects a subcommand; one of the
    following:

    * ``rattail mailmon start``
    * ``rattail mailmon stop``
    """
    name = 'mailmon'
    description = "Manage the mail monitor daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')

        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/mailmon.pid',
                            help="Path to PID file.")

    def run(self, args):
        from rattail.mailmon.daemon import MailMonitorDaemon

        daemon = MailMonitorDaemon(args.pidfile, config=self.config)
        if args.subcommand == 'stop':
            daemon.stop()
        else: # start
            try:
                daemon.start(daemonize=False)
            except KeyboardInterrupt:
                self.stderr.write("Interrupted.\n")


class MakeAppDir(Subcommand):
    """
    Create a conventional 'app' dir for a virtual environment
    """
    name = 'make-appdir'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--path', metavar='PATH',
                            help="Path where the app folder is to be established.  If not "
                            "specified, a default will be assumed based on the virtual "
                            "environment, e.g. '/envs/rattail/app'.")
        parser.add_argument('-U', '--user', metavar='USERNAME',
                            help="Linux username which should be given ownership to the various "
                            "data folders which are to be created.  This is used when the app(s) "
                            "are to normally be ran as the 'rattail' user for instance.  Use "
                            "of this option probably requires 'sudo' or equivalent.")

    def run(self, args):
        if args.path:
            app_path = os.path.abspath(args.path)
            if os.path.basename(app_path) != 'app':
                app_path = os.path.join(app_path, 'app')
        else:
            app_path = os.path.join(sys.prefix, 'app')
        app_path = app_path.rstrip('/')

        self.app.make_appdir(app_path)

        # TODO: this bit should probably be part of app method
        if args.user:
            import pwd
            pwdata = pwd.getpwnam(args.user)
            folders = [
                'data',
                os.path.join('data', 'uploads'),
                'log',
                'sessions',
                'work',
            ]
            for name in folders:
                path = os.path.join(app_path, name)
                os.chown(path, pwdata.pw_uid, pwdata.pw_gid)

        self.stdout.write("Rattail app dir generated at: {}\n".format(app_path))


class MakeConfig(Subcommand):
    """
    Generate stub config file(s) where you want them
    """
    name = 'make-config'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('-T', '--type', metavar='NAME', default='rattail',
                            help="Type of config file to create; defaults to 'rattail' "
                            "which will generate 'rattail.conf'")
        parser.add_argument('-O', '--output', metavar='PATH', default='.',
                            help="Path where the config file is to be generated.  This can "
                            "be the full path including filename, or just the folder, in which "
                            "case the filename is inferred from 'type'.  Default is to current "
                            "working folder.")
        parser.add_argument('--list-types', action='store_true',
                            help="List the types of config files this tool can "
                            "generate.")

    def run(self, args):
        if args.list_types:
            self.list_types()
            return

        template_path = self.app.find_config_template(args.type)
        if not template_path:
            self.stderr.write("config template not found for type: {}\n".format(args.type))
            sys.exit(1)

        output_path = os.path.abspath(args.output)
        if os.path.isdir(output_path):
            output_path = os.path.join(output_path, '{}.conf'.format(args.type))
        if os.path.exists(output_path):
            self.stderr.write("ERROR! output file already exists: {}\n".format(output_path))
            sys.exit(2)

        config_path = self.app.make_config_file(args.type, output_path,
                                                template_path=template_path)
        self.stdout.write("Config file generated at: {}\n".format(config_path))

    def list_types(self):
        templates = self.app.get_all_config_templates()
        self.stdout.write("CONFIG TEMPLATES:\n")
        self.stdout.write("=========================\n")
        for name, path in templates.items():
            self.stdout.write("{:25s} {}\n".format(name, path))
        self.stdout.write("=========================\n")


class MakeUser(Subcommand):
    """
    Create a new user account in a given system
    """
    name = 'make-user'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('username',
                            help="Username for the new user.")
        parser.add_argument('--system', default='rattail',
                            help="System in which to create the new user; defaults to "
                            "rattail; must be one of: rattail, windows")
        parser.add_argument('-A', '--admin', action='store_true',
                            help="Whether the new user should have admin rights within "
                            "the system (if applicable).")
        parser.add_argument('--password',
                            help="Password to set for the new user.  If not specified, "
                            "you may be prompted for one.")
        parser.add_argument('--no-password', action='store_true',
                            help="Do not ask for, or try to set, a password for the new user.")
        parser.add_argument('--full-name',
                            help="Full (display) name for the new user (if applicable).")
        parser.add_argument('--comment',
                            help="Comment string for the new user (if applicable).")
        parser.add_argument('--groups',
                            help="Optional list of groups (roles) to which the new user "
                            "should be assigned.")

    def run(self, args):
        mkuser = getattr(self, 'mkuser_{}'.format(args.system), None)
        if mkuser:
            if mkuser(args):
                self.stdout.write("created new user in '{}' system: {}\n".format(
                    args.system, args.username))
        else:
            self.stderr.write("don't know how to make user for '{}' system\n".format(args.system))
            sys.exit(1)

    def user_exists(self, args):
        self.stdout.write("user already exists in '{}' system: {}\n".format(
            args.system, args.username))
        sys.exit(1)

    def obtain_password(self, args):
        if args.password:
            return args.password
        try:
            password = None
            while not password:
                password = getpass(str("enter password for new user: ")).strip()
        except KeyboardInterrupt:
            self.stderr.write("\noperation canceled by user\n")
            sys.exit(2)
        return password

    def mkuser_rattail(self, args):
        from sqlalchemy import orm
        from rattail.db import auth

        session = self.make_session()
        model = self.parent.db_model
        if session.query(model.User).filter_by(username=args.username).count():
            session.close()
            return self.user_exists(args)

        roles = []
        if args.groups:
            for name in parse_list(args.groups):
                try:
                    role = session.query(model.Role)\
                                  .filter(model.Role.name == name)\
                                  .one()
                except orm.exc.NoResultFound:
                    self.stderr.write("Role not found: {}\n".format(name))
                    session.close()
                    sys.exit(4)
                else:
                    roles.append(role)

        user = model.User(username=args.username)
        if not args.no_password:
            auth.set_user_password(user, self.obtain_password(args))

        if args.admin:
            user.roles.append(auth.administrator_role(session))
        for role in roles:
            user.roles.append(role)

        if args.full_name:
            kwargs = {'display_name': args.full_name}
            words = args.full_name.split()
            if len(words) == 2:
                kwargs.update({'first_name': words[0], 'last_name': words[1]})
            user.person = model.Person(**kwargs)

        session.add(user)
        session.commit()
        session.close()
        return True

    def mkuser_windows(self, args):
        if sys.platform != 'win32':
            self.stderr.write("sorry, only win32 platform is supported\n")
            sys.exit(1)

        from rattail.win32 import users
        from rattail.win32 import require_elevation

        if args.no_password:
            self.stderr.write("sorry, a password is required when making a 'win32' user\n")
            sys.exit(1)

        require_elevation()

        if users.user_exists(args.username):
            return self.user_exists(args)

        return users.create_user(args.username, self.obtain_password(args),
                                 full_name=args.full_name, comment=args.comment)


class MakeUUID(Subcommand):
    """
    Generate a new UUID
    """
    name = 'make-uuid'
    description = __doc__.strip()

    def run(self, args):
        from rattail.core import get_uuid

        self.stdout.write("{}\n".format(get_uuid()))


class PalmCommand(Subcommand):
    """
    Manages registration for the HotSync Manager conduit; called as::

       rattail palm
    """

    name = 'palm'
    description = "Manage the HotSync Manager conduit registration"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        register = subparsers.add_parser('register', help="Register Rattail conduit")
        register.set_defaults(subcommand='register')

        unregister = subparsers.add_parser('unregister', help="Unregister Rattail conduit")
        unregister.set_defaults(subcommand='unregister')

    def run(self, args):
        from rattail import palm
        from rattail.win32 import require_elevation
        from rattail.exceptions import PalmError

        require_elevation()

        if args.subcommand == 'register':
            try:
                palm.register_conduit()
            except PalmError as error:
                sys.stderr.write("{}\n".format(error))

        elif args.subcommand == 'unregister':
            try:
                palm.unregister_conduit()
            except PalmError as error:
                sys.stderr.write("{}\n".format(error))
                

class RunAndMail(Subcommand):
    """
    Run a command as subprocess, and email the result/output
    """
    name = 'run-n-mail'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--skip-if-empty', action='store_true',
                            help="Skip sending the email if the command generates no output.")

        parser.add_argument('--key', default='run_n_mail',
                            help="Config key for email settings")
        # TODO: these all seem like good ideas, but not needed yet?
        # parser.add_argument('--from', '-F', metavar='ADDRESS',
        #                     help="Override value of From: header")
        # parser.add_argument('--to', '-T', metavar='ADDRESS',
        #                     help="Override value of To: header (may specify more than once)")
        # parser.add_argument('--cc', metavar='ADDRESS',
        #                     help="Override value of Cc: header (may specify more than once)")
        # parser.add_argument('--bcc', metavar='ADDRESS',
        #                     help="Override value of Bcc: header (may specify more than once)")
        parser.add_argument('--subject', '-S',
                            help="Override value of Subject: header (i.e. value after prefix)")
        parser.add_argument('cmd', metavar='COMMAND',
                            help="Command which should be ran, and result of which will be emailed")

        parser.add_argument('--keep-exit-code', action='store_true',
                            help="Exit with same return code as subprocess.  If "
                            "this is not specified, `run-n-mail` will normally "
                            "exit with code 0 regardless of what happens with "
                            "the subprocess.")

    def run(self, args):
        from rattail.mail import send_email

        cmd = parse_list(args.cmd)
        log.info("will run command as subprocess: %s", cmd)
        run_began = make_utc()

        try:
            # TODO: must we allow for shell=True in some situations? (clearly not yet)
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            retcode = 0
            log.info("command completed successfully")
        except subprocess.CalledProcessError as error:
            output = error.output
            retcode = error.returncode
            log.warning("command exited with code: %s", retcode)

        output = output.decode(errors='replace')

        run_ended = make_utc()
        runtime = run_ended - run_began
        runtime_pretty = humanize.naturaldelta(runtime)

        if args.skip_if_empty and not output:
            log.info("command had no output, so will skip sending email")
            # return

        else: # send email
            kwargs = {}
            if args.subject:
                kwargs['subject_template'] = args.subject
            send_email(self.config, args.key, {
                'cmd': cmd,
                'run_began': run_began,
                'run_ended': run_ended,
                'runtime': runtime,
                'runtime_pretty': runtime_pretty,
                'retcode': retcode,
                'output': output,
            }, **kwargs)

        if retcode and args.keep_exit_code:
            sys.exit(retcode)


class RunSQL(Subcommand):
    """
    Run (first statement of) a SQL script against a database
    """
    name = 'runsql'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('engine',
                            help="SQLAlchemy engine URL for the database.")
        parser.add_argument('script', type=argparse.FileType('r'),
                            help="Path to file which contains a SQL script.")
        parser.add_argument('--max-width', type=int, default=80,
                            help="Max table width when displaying results.")

    def run(self, args):
        import sqlalchemy as sa
        import texttable

        sql = []
        for line in args.script:
            line = line.strip()
            if line and not line.startswith('--'):
                sql.append(line)
                if line.endswith(';'):
                    break

        sql = ' '.join(sql)
        engine = sa.create_engine(args.engine)

        with engine.begin() as cxn:
            result = cxn.execute(sa.text(sql))
            rows = result.fetchall()
            if rows:
                table = texttable.Texttable(max_width=args.max_width)

                # force all columns to be treated as text.  that seems a bit
                # heavy-handed but is definitely the simplest way to deal with
                # issues such as a UPC being displayed in scientific notation
                table.set_cols_dtype(['t' for col in rows[0]])

                # add a header row, plus all data rows
                table.add_rows([rows[0].keys()] + rows)

                self.stdout.write("{}\n".format(table.draw()))


class Upgrade(Subcommand):
    """
    Upgrade the local Rattail app
    """
    name = 'upgrade'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('--system', default='rattail',
                            help="System to which the upgrade applies.")

        parser.add_argument('--description',
                            help="Description for the new/matched upgrade.")
        parser.add_argument('--enabled', action='store_true', default=True,
                            help="Indicate the enabled flag should be ON for the new/matched upgrade.  "
                            "Note that this is the default if you do not specify.")
        parser.add_argument('--no-enabled', action='store_false', dest='enabled',
                            help="Indicate the enabled flag should be OFF for the new/matched upgrade.")
        parser.add_argument('--create', action='store_true',
                            help="Create a new upgrade with the given attributes.")
        parser.add_argument('--execute', action='store_true',
                            help="Execute the upgrade.  Note that if you do not specify "
                            "--create then the upgrade matching the given attributes "
                            "will be read from the database.  If such an upgrade is not "
                            "found or is otherwise invalid (e.g. already executed), "
                            "the command will fail.")

        parser.add_argument('--keep-exit-code', action='store_true',
                            help="Exit with same return code as subprocess.  If "
                            "this is not specified, this command will normally "
                            "exit with code 0 regardless of what happens with "
                            "the subprocess.  (only applicable with --execute)")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        from sqlalchemy import orm

        if not args.create and not args.execute:
            self.stderr.write("Must specify --create and/or --execute\n")
            sys.exit(1)

        session = self.make_session()
        model = self.model
        user = self.get_runas_user(session)

        if args.create:
            upgrade = model.Upgrade()
            upgrade.system = args.system or 'rattail'
            upgrade.description = args.description
            upgrade.created = make_utc()
            upgrade.created_by = user
            upgrade.enabled = args.enabled
            session.add(upgrade)
            session.flush()
            log.info("user '%s' created new upgrade: %s", user.username, upgrade)

        else:
            upgrades = session.query(model.Upgrade)\
                              .filter(model.Upgrade.enabled == args.enabled)
            if args.description:
                upgrades = upgrades.filter(model.Upgrade.description == args.description)
            try:
                upgrade = upgrades.one()
            except orm.exc.NoResultFound:
                self.stderr.write("no matching upgrade found\n")
                session.rollback()
                session.close()
                sys.exit(1)
            except orm.exc.MultipleResultsFound:
                self.stderr.write("found {} matching upgrades\n".format(upgrades.count()))
                session.rollback()
                session.close()
                sys.exit(1)

        if args.execute:
            if upgrade.executed:
                self.stderr.write("upgrade has already been executed: {}\n".format(upgrade))
                session.rollback()
                session.close()
                sys.exit(1)
            if not upgrade.enabled:
                self.stderr.write("upgrade is not enabled for execution: {}\n".format(upgrade))
                session.rollback()
                session.close()
                sys.exit(1)

            # execute upgrade
            handler = self.app.get_upgrade_handler()
            log.info("will now execute upgrade: %s", upgrade)
            if not args.dry_run:
                handler.mark_executing(upgrade)
                session.commit()
                handler.do_execute(upgrade, user, progress=self.progress)
            log.info("user '%s' executed upgrade: %s", user.username, upgrade)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

        if (args.execute and not args.dry_run
            and args.keep_exit_code and upgrade.exit_code):
            sys.exit(upgrade.exit_code)


def main(*args):
    """
    The primary entry point for the Rattail command system.
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)



# -*- coding: utf-8 -*-
"""add customer.active_in_pos_sticky

Revision ID: cc3de1ca0689
Revises: a9b37b388e56
Create Date: 2017-08-04 11:06:15.835698

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'cc3de1ca0689'
down_revision = u'a9b37b388e56'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # customer
    op.add_column('customer', sa.Column('active_in_pos_sticky', sa.Boolean(), nullable=True))
    op.add_column('customer_version', sa.Column('active_in_pos_sticky', sa.Boolean(), autoincrement=False, nullable=True))
    customer = sa.sql.table('customer', sa.sql.column('active_in_pos_sticky'))
    op.execute(customer.update().values({'active_in_pos_sticky': False}))
    op.alter_column('customer', 'active_in_pos_sticky', nullable=False,
                    existing_type=sa.Boolean())


def downgrade():

    # customer
    op.drop_column('customer_version', 'active_in_pos_sticky')
    op.drop_column('customer', 'active_in_pos_sticky')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Employees
"""

import datetime

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

from .core import Base, uuid_column, getset_factory
from .people import Person
from .contact import PhoneNumber, EmailAddress
from .stores import Store
from .org import Department


class Employee(Base):
    """
    Represents an employee within the organization.
    """
    __tablename__ = 'employee'
    __table_args__ = (
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='employee_fk_person'),
        sa.UniqueConstraint('person_uuid', name='employee_uq_person'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    id = sa.Column(sa.Integer(), nullable=True, doc="""
    Numeric ID for the employee, if applicable/known.
    """)

    person_uuid = sa.Column(sa.String(length=32), nullable=False)
    person = orm.relationship(Person, back_populates='employee', doc="""
    Reference to the person to whom this employee record pertains.
    """)

    status = sa.Column(sa.Integer(), nullable=True, doc="""
    Status code for the employee, e.g. current vs. former.
    """)

    display_name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Display name for the employee.  Whereas the *person's* display name is
    generally their full name, this *employee* display name is often an
    abbreviated version which is visible on printed receipts, etc.
    """)

    full_time = sa.Column(sa.Boolean(), nullable=True, doc="""
    Simple flag to indicate if employee is considered full-time.  This flag
    should theoretically only be relevant if employee is also *current*.
    """)

    full_time_start = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which employee became full time, if known/relevant.
    """)

    first_name = association_proxy('person', 'first_name',
                                   getset_factory=getset_factory,
                                   creator=lambda n: Person(first_name=n))
    last_name = association_proxy('person', 'last_name',
                                  getset_factory=getset_factory,
                                  creator=lambda n: Person(last_name=n))
    users = association_proxy('person', 'users')

    @property
    def user(self):
        return self.users[0] if self.users else None

    def __str__(self):
        return str(str(self.person or "") or self.display_name or "")

    @property
    def customers(self):
        return self.person.customers

    def only_customer(self, require=True):
        """
        Convenience method which invokes ``self.person.only_customer()``
        """
        return self.person.only_customer(require=require)

    def add_email_address(self, address, type='Home'):
        email = EmployeeEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = EmployeePhoneNumber(number=number, type=type)
        self.phones.append(phone)

    def sorted_history(self, reverse=False):
        """
        Returns the employee history, sorted in a "reasonable" way.
        """
        old_date = datetime.date(1900, 1, 1)
        new_date = datetime.date.today() + datetime.timedelta(days=7)
        return sorted(self.history,
                      key=lambda h: (h.start_date or old_date, h.end_date or new_date),
                      reverse=reverse)

    def get_current_history(self):
        """
        Returns the "current" history record for the employee, if applicable.
        Note that this history record is not necessarily "active" - it's just
        the most recent.
        """
        if self.history:
            history = self.sorted_history(reverse=True)
            return history[0]


Person.employee = orm.relationship(
    Employee,
    back_populates='person',
    uselist=False,
    doc="""
    Reference to the :class:`Employee` record for the person, if there is one.
    """)


class EmployeePhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Employee`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Employee'}


Employee.phones = orm.relationship(
    EmployeePhoneNumber,
    backref='employee',
    primaryjoin=EmployeePhoneNumber.parent_uuid == Employee.uuid,
    foreign_keys=[EmployeePhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=EmployeePhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Employee.phone = orm.relationship(
    EmployeePhoneNumber,
    primaryjoin=sa.and_(
        EmployeePhoneNumber.parent_uuid == Employee.uuid,
        EmployeePhoneNumber.preference == 1),
    foreign_keys=[EmployeePhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class EmployeeEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Employee`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Employee'}


Employee.emails = orm.relationship(
    EmployeeEmailAddress,
    backref='employee',
    primaryjoin=EmployeeEmailAddress.parent_uuid == Employee.uuid,
    foreign_keys=[EmployeeEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=EmployeeEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Employee.email = orm.relationship(
    EmployeeEmailAddress,
    primaryjoin=sa.and_(
        EmployeeEmailAddress.parent_uuid == Employee.uuid,
        EmployeeEmailAddress.preference == 1),
    foreign_keys=[EmployeeEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)


class EmployeeStore(Base):
    """
    Represents the association between an employee and a store.
    """
    __tablename__ = 'employee_x_store'
    __table_args__ = (
        sa.ForeignKeyConstraint(['employee_uuid'], ['employee.uuid'], name='employee_x_store_fk_employee'),
        sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'], name='employee_x_store_fk_store'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    employee_uuid = sa.Column(sa.String(length=32), nullable=False)
    store_uuid = sa.Column(sa.String(length=32), nullable=False)

    employee = orm.relationship(
        Employee,
        backref=orm.backref(
            '_stores',
            cascade='all, delete-orphan',
            cascade_backrefs=False))

    store = orm.relationship(
        Store,
        order_by=Store.name,
        backref=orm.backref(
            '_employees',
            cascade='all, delete-orphan',
            cascade_backrefs=False))

Employee.stores = association_proxy(
    '_stores', 'store',
    creator=lambda x: EmployeeStore(store=x),
    getset_factory=getset_factory)

Store.employees = association_proxy(
    '_employees', 'employee',
    creator=lambda x: EmployeeStore(employee=x),
    getset_factory=getset_factory)


class EmployeeDepartment(Base):
    """
    Represents the association between an employee and a department.
    """
    __tablename__ = 'employee_x_department'
    __table_args__ = (
        sa.ForeignKeyConstraint(['employee_uuid'], ['employee.uuid'], name='employee_x_department_fk_employee'),
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='employee_x_department_fk_department'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    employee_uuid = sa.Column(sa.String(length=32), nullable=False)
    department_uuid = sa.Column(sa.String(length=32), nullable=False)

    employee = orm.relationship(
        Employee,
        backref=orm.backref(
            '_departments',
            cascade='all, delete-orphan',
            cascade_backrefs=False))

    department = orm.relationship(
        Department,
        order_by=Department.name,
        backref=orm.backref(
            '_employees',
            cascade='all, delete-orphan',
            cascade_backrefs=False))

Employee.departments = association_proxy(
    '_departments', 'department',
    creator=lambda x: EmployeeDepartment(department=x),
    getset_factory=getset_factory)

Department.employees = association_proxy(
    '_employees', 'employee',
    creator=lambda x: EmployeeDepartment(employee=x),
    getset_factory=getset_factory)


class EmployeeHistory(Base):
    """
    Represents a period of time for which a person was employed by the organization.
    """
    __tablename__ = 'employee_history'
    __table_args__ = (
        sa.ForeignKeyConstraint(['employee_uuid'], ['employee.uuid'], name='employee_history_fk_employee'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    employee_uuid = sa.Column(sa.String(length=32), nullable=False)
    employee = orm.relationship(
        Employee, doc="""
        Reference to the employee to which this history record pertains.
        """,
        backref=orm.backref(
            'history',
            order_by='EmployeeHistory.start_date',
            cascade_backrefs=False,
            doc="""
            Sequence of history records for the employee.
            """))

    start_date = sa.Column(sa.Date(), nullable=False, doc="""
    Date on which the employee became active in the organization.
    """)

    end_date = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the employee became *inactive*.
    """)



# -*- coding: utf-8; -*-
"""add allowances to vendorcatalog batch

Revision ID: 727c63fcde2d
Revises: 4e5526fae79e
Create Date: 2019-04-18 16:16:44.844542

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '727c63fcde2d'
down_revision = '4e5526fae79e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_catalog_row
    op.add_column('vendor_catalog_row', sa.Column('discount_amount', sa.Numeric(precision=9, scale=5), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('discount_ends', sa.DateTime(), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('discount_percent', sa.Numeric(precision=7, scale=4), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('discount_starts', sa.DateTime(), nullable=True))


def downgrade():

    # vendor_catalog_row
    op.drop_column('vendor_catalog_row', 'discount_starts')
    op.drop_column('vendor_catalog_row', 'discount_percent')
    op.drop_column('vendor_catalog_row', 'discount_ends')
    op.drop_column('vendor_catalog_row', 'discount_amount')



# -*- coding: utf-8 -*-
"""add more newproduct batch row attrs

Revision ID: a20208a41889
Revises: 3b70d683b0cf
Create Date: 2022-08-30 13:44:05.579769

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'a20208a41889'
down_revision = '3b70d683b0cf'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_newproduct_row
    op.add_column('batch_newproduct_row', sa.Column('weighed', sa.Boolean(), nullable=True))
    op.add_column('batch_newproduct_row', sa.Column('tax1', sa.Boolean(), nullable=True))
    op.add_column('batch_newproduct_row', sa.Column('tax2', sa.Boolean(), nullable=True))
    op.add_column('batch_newproduct_row', sa.Column('tax3', sa.Boolean(), nullable=True))
    op.add_column('batch_newproduct_row', sa.Column('ecommerce_available', sa.Boolean(), nullable=True))


def downgrade():

    # batch_newproduct_row
    op.drop_column('batch_newproduct_row', 'ecommerce_available')
    op.drop_column('batch_newproduct_row', 'tax3')
    op.drop_column('batch_newproduct_row', 'tax2')
    op.drop_column('batch_newproduct_row', 'tax1')
    op.drop_column('batch_newproduct_row', 'weighed')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Poser Handler
"""

import importlib
import os
import sys
import subprocess
import logging
from collections import OrderedDict

from mako.lookup import TemplateLookup

from rattail.app import GenericHandler
from rattail.files import resource_path
from rattail.util import simple_error
from rattail.reporting import ExcelReport
from rattail.config import ConfigExtension


log = logging.getLogger(__name__)


class PoserHandler(GenericHandler):
    """
    Base class and default implementation for Poser (custom code)
    handler.
    """

    def sanity_check(self):
        poserdir = self.get_default_poser_dir()
        return os.path.exists(poserdir)

    def get_default_poser_dir(self):
        appdir = self.config.appdir(require=False)
        if not appdir:
            appdir = os.path.join(sys.prefix, 'app')
        return os.path.join(appdir, 'poser')

    def make_poser_dir(self, path=None, **kwargs):
        """
        Create the directory structure for Poser.
        """
        # assume default path if none specified
        if not path:
            path = self.get_default_poser_dir()

        # path must not yet exist
        path = os.path.abspath(path)
        if os.path.exists(path):
            raise RuntimeError("folder already exists: {}".format(path))

        # make top-level dir
        os.makedirs(path)

        # normal refresh takes care of most of it
        self.refresh_poser_dir(path)

        # make git repo
        subprocess.check_call(['git', 'init', path])
        subprocess.check_call([
            'bash', '-c',
            "cd {} && git add poser .gitignore".format(path),
        ])

        return path

    def refresh_poser_dir(self, path=None, **kwargs):
        """
        Refresh the basic structure for Poser.
        """
        # assume default path if none specified
        if not path:
            path = self.get_default_poser_dir()

        # path must already exist
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise RuntimeError("folder does not exist: {}".format(path))

        # make poser pkg dir
        poser = os.path.join(path, 'poser')
        if not os.path.exists(poser):
            os.makedirs(poser)

        # add `__init__` stub
        init = os.path.join(poser, '__init__.py')
        if not os.path.exists(init):
            with open(init, 'wt') as f:
                pass

        # make 'db' subpackage
        db = os.path.join(poser, 'db')
        if not os.path.exists(db):
            os.makedirs(db)

        # add `__init__` stub
        init = os.path.join(db, '__init__.py')
        if not os.path.exists(init):
            with open(init, 'wt') as f:
                pass

        # make 'db.model' subpackage
        model = os.path.join(db, 'model')
        if not os.path.exists(model):
            os.makedirs(model)

        # add `__init__` stub
        init = os.path.join(model, '__init__.py')
        if not os.path.exists(init):
            with open(init, 'wt') as f:
                pass

        # make 'db:alembic' folder
        alembic = os.path.join(db, 'alembic')
        if not os.path.exists(alembic):
            os.makedirs(alembic)

        # make 'reports' subpackage
        reports = os.path.join(poser, 'reports')
        if not os.path.exists(reports):
            os.makedirs(reports)
        init = os.path.join(reports, '__init__.py')
        if not os.path.exists(init):
            with open(init, 'wt') as f:
                pass

        # make 'web' subpackage
        web = os.path.join(poser, 'web')
        if not os.path.exists(web):
            os.makedirs(web)
        init = os.path.join(web, '__init__.py')
        if not os.path.exists(init):
            with open(init, 'wt') as f:
                pass

        # make 'web.views' subpackage
        views = os.path.join(poser, 'web', 'views')
        if not os.path.exists(views):
            os.makedirs(views)
        init = os.path.join(views, '__init__.py')
        if not os.path.exists(init):
            with open(init, 'wt') as f:
                pass

        # make .gitignore
        gitignore = os.path.join(path, '.gitignore')
        # TODO: this should always overwrite a "managed" section of the file
        if not os.path.exists(gitignore):
            with open(gitignore, 'wt') as f:
                f.write('**/__pycache__/\n')

    def get_supported_report_flavors(self, **kwargs):
        # TODO: these should come from entry points etc.
        flavors = {
            'customer_py': {
                'description': "Query the Customer table (Python)",
                'template': '/reports/customer_py.mako',
                'output_fields': ['customer_number', 'customer_name'],
            },
            'customer_sql': {
                'description': "Query the Customer table (SQL)",
                'template': '/reports/customer_sql.mako',
                'output_fields': ['customer_number', 'customer_name'],
            },
            'product_py': {
                'description': "Query the Product table (Python)",
                'template': '/reports/product_py.mako',
                'output_fields': ['product_upc', 'product_description'],
            },
            'product_sql': {
                'description': "Query the Product table (SQL)",
                'template': '/reports/product_sql.mako',
                'output_fields': ['product_upc', 'product_description'],
            },
        }

        items = sorted(flavors.items(),
                       key=lambda itm: itm[1]['description'])

        items.insert(0, ('default', {
            'description': "Default (empty)",
            'template': '/reports/base.mako',
            'output_fields': [],
        }))

        return OrderedDict(items)

    def get_all_reports(self, ignore_errors=True, **kwargs):
        try:
            from poser import reports
        except:
            if ignore_errors:
                log.warning("cannot import poser reports module!")
                return []
            raise

        all_reports = []

        path = os.path.dirname(reports.__file__)
        for fname in os.listdir(path):
            if fname.startswith('_'):
                continue
            if not fname.endswith('.py'):
                continue

            key, ext = os.path.splitext(
                os.path.basename(fname))

            report = self.normalize_report(key)
            if report:
                all_reports.append(report)

        return all_reports

    def normalize_report(self, key, **kwargs):
        """
        Load the report module for the given key, and normalize the
        report it contains.
        """
        from poser import reports

        filepath = os.path.join(os.path.dirname(reports.__file__),
                                '{}.py'.format(key))

        modpath = '.'.join((reports.__name__, key))

        try:
            mod = importlib.import_module(modpath)
            importlib.reload(mod)

        except Exception as error:
            log.warning("import failed for %s", modpath, exc_info=True)
            return {
                'report_key': key,
                'module_file_path': filepath,
                'error': simple_error(error),
                'report': None,
                'report_name': '??',
                'description': '??',
            }

        for name in dir(mod):
            if name.startswith('_'):
                continue

            obj = getattr(mod, name)
            if (isinstance(obj, type)
                and issubclass(obj, ExcelReport)
                and obj.type_key
                and obj.type_key.startswith('poser_')):

                return {
                    'report': obj,
                    'module_file_path': filepath,
                    'report_key': key,
                    'report_name': obj.name,
                    'description': obj.__doc__,
                }

    def normalize_report_path(self, path, **kwargs):
        """
        Normalize the report module for the given file path.
        """
        from poser import reports

        # is module in the poser reports package?
        dirname = os.path.dirname(path)
        if dirname != os.path.dirname(reports.__file__):
            return

        # does module have a "normal" name?
        fname = os.path.basename(path)
        if fname.startswith('_'):
            return
        if not fname.endswith('.py'):
            return

        # okay then normalize per usual
        key, ext = os.path.splitext(fname)
        return self.normalize_report(key)

    def make_report(self, key, name, description,
                    flavor=None, output_fields=[], include_comments=True,
                    **kwargs):
        """
        Generate a new Python module for a Poser Report.
        """
        from poser import reports

        # TODO: make templates dir configurable
        templates = [resource_path('rattail:templates/poser')]
        templates = TemplateLookup(directories=templates)

        output_fields = output_fields or []
        template = '/reports/base.mako'
        if flavor:
            flavors = self.get_supported_report_flavors()
            if flavor in flavors:
                template = flavors[flavor]['template']
                output_fields = flavors[flavor]['output_fields']

        template = templates.get_template(template)

        cap_name = ''.join([word.capitalize()
                            for word in key.split('_')])

        context = {
            'app_title': self.config.app_title(),
            'key': key,
            'name': name,
            'cap_name': cap_name,
            'description': description,
            'output_fields': output_fields,
            'include_comments': include_comments,
        }

        path = os.path.join(os.path.dirname(reports.__file__),
                            '{}.py'.format(key))

        with open(path, 'wt') as f:
            f.write(template.render(**context))

        return self.normalize_report(key)

    def replace_report(self, key, path, **kwargs):
        """
        Replace the report module from the given file path.
        """
        from poser import reports

        report = self.normalize_report(key)
        oldpath = report['module_file_path']
        newpath = os.path.join(os.path.dirname(reports.__file__),
                               os.path.basename(path))

        if newpath != oldpath and os.path.exists(newpath):
            raise RuntimeError("Report already exists; cannot overwrite: {}".format(newpath))

        if os.path.exists(oldpath):
            os.remove(oldpath)

        with open(path, 'rb') as fin:
            with open(newpath, 'wb') as fout:
                fout.write(fin.read())

        return self.normalize_report_path(newpath)

    def delete_report(self, key, **kwargs):
        from poser import reports

        path = os.path.join(os.path.dirname(reports.__file__),
                            '{}.py'.format(key))
        if os.path.exists(path):
            os.remove(path)

    def get_all_tailbone_views(self, **kwargs):
        from poser.web import views

        all_views = []

        path = os.path.dirname(views.__file__)
        for fname in os.listdir(path):
            if fname.startswith('_'):
                continue
            if not fname.endswith('.py'):
                continue

            key, ext = os.path.splitext(
                os.path.basename(fname))

            view = self.normalize_view(key)
            if view:
                all_views.append(view)

        return all_views

    def normalize_tailbone_view(self, key, **kwargs):
        """
        Load the view module for the given key, and normalize the view
        it contains.
        """
        from tailbone.views import MasterView
        from poser.web import views

        filepath = os.path.join(os.path.dirname(views.__file__),
                                '{}.py'.format(key))

        modpath = '.'.join((views.__name__, key))

        try:
            mod = importlib.import_module(modpath)
            importlib.reload(mod)

        except Exception as error:
            log.warning("import failed for %s", modpath, exc_info=True)
            return {
                'key': key,
                'module_file_path': filepath,
                'error': simple_error(error),
                'view': None,
                'class_name': '??',
                'description': '??',
            }

        for name in dir(mod):
            if name.startswith('_'):
                continue

            obj = getattr(mod, name)
            if (isinstance(obj, type)
                and issubclass(obj, MasterView)
                and hasattr(obj, 'route_prefix')
                and obj.route_prefix.startswith('poser.')):

                return {
                    'key': key,
                    'module_file_path': filepath,
                    'view': obj,
                    'class_name': obj.__name__,
                    'description': obj.__doc__,
                }



# -*- coding: utf-8; -*-
"""add pricing_batch.suggested_price

Revision ID: e645676eb37a
Revises: 5ef33e961026
Create Date: 2018-11-18 19:54:55.006187

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'e645676eb37a'
down_revision = '5ef33e961026'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing_row
    op.add_column('batch_pricing_row', sa.Column('suggested_price', sa.Numeric(precision=8, scale=3), nullable=True))


def downgrade():

    # batch_pricing_row
    op.drop_column('batch_pricing_row', 'suggested_price')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data models for scheduled app upgrades
"""

import datetime

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db.model import Base, uuid_column, User


class Upgrade(Base):
    """
    Represents a scheduled app upgrade.
    """
    __tablename__ = 'upgrade'
    __table_args__ = (
        sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='upgrade_fk_created_by'),
        sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'], name='upgrade_fk_executed_by'),
    )

    uuid = uuid_column()

    system = sa.Column(sa.String(length=100), nullable=True, doc="""
    System to which the upgrade applies.  Empty values here usually
    mean the ugprade applies to the primary Rattail app.
    """)

    created = sa.Column(sa.DateTime(), nullable=True, default=datetime.datetime.utcnow, doc="""
    Date and time when the upgrade record was created.
    """)

    created_by_uuid = sa.Column(sa.String(length=32), nullable=False)
    created_by = orm.relationship(
        User,
        foreign_keys=[created_by_uuid],
        doc="""
        Reference to user who created the upgrade record.
        """)

    description = sa.Column(sa.String(length=255), nullable=False, doc="""
    Basic (identifying) description for the upgrade.
    """)

    not_until = sa.Column(sa.DateTime(), nullable=True, doc="""
    If this is null, upgrade may occur at any time.  Otherwise the upgrade
    should not occur until after this time.
    """)

    notes = sa.Column(sa.Text(), nullable=True, doc="""
    Generic notes for the upgrade.
    """)

    enabled = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Whether or not the upgrade should be allowed to occur at all.
    """)

    executing = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Whether or not the upgrade is currently being performed.
    """)

    status_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Current status code for the upgrade.
    """)

    executed = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the upgrade was performed.
    """)

    executed_by_uuid = sa.Column(sa.String(length=32), nullable=True)
    executed_by = orm.relationship(
        User,
        foreign_keys=[executed_by_uuid],
        doc="""
        Reference to user who performed the upgrade.
        """)

    exit_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Exit code for the upgrade execution process, if applicable.
    """)

    def __str__(self):
        return str(self.description or "")


class UpgradeRequirement(Base):
    """
    A package version requirement for an app upgrade.
    """
    __tablename__ = 'upgrade_requirement'
    __table_args__ = (
        sa.ForeignKeyConstraint(['upgrade_uuid'], ['upgrade.uuid'], name='upgrade_requirement_fk_upgrade'),
    )

    uuid = uuid_column()

    upgrade_uuid = sa.Column(sa.String(length=32), nullable=False)
    upgrade = orm.relationship(
        Upgrade,
        backref=orm.backref(
            'requirements',
            cascade_backrefs=False))

    package = sa.Column(sa.String(length=255), nullable=False, doc="""
    Package name as found on PyPI (or at least interpretable by pip).
    """)

    version = sa.Column(sa.String(length=255), nullable=False, default='LATEST', doc="""
    Version requirement for the package.  If the value is ``'LATEST'`` then the
    latest available version will be installed; otherwise the value is applied
    as-is to the package when building the requirement string.
    """)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Backup commands
"""

import os
import re
import shutil
import socket
import subprocess
import sys
import logging

from rattail.commands.core import Subcommand
from rattail.config import parse_bool, parse_list


log = logging.getLogger(__name__)


class Backup(Subcommand):
    """
    Backup the database(s) and/or files for this machine
    """
    name = 'backup'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--list-dbs', action='store_true',
                            help="List all databases which may be backed up, then exit.")

        parser.add_argument('--dbdump', action='store_true',
                            help="Dump some or all databases during the backup run.  Exactly "
                            "which databases are dumped, will depend on other parameters.")
        parser.add_argument('--no-dbdump', action='store_true',
                            help="Do NOT dump any databases during the backup run.")

        parser.add_argument('--dbdump-include', action='append', dest='included_dbs', metavar='NAME',
                            help="Name of database to include in the dump.  May be specified "
                            "more than once.  Note that if this parameter is used, then ONLY "
                            "those databases specified will be backed up. Also, the --dbdump-exclude "
                            "parameter(s) will be ignored.")
        parser.add_argument('--dbdump-exclude', action='append', dest='excluded_dbs', metavar='NAME',
                            help="Name of database to exclude from the dump.  May be specified "
                            "more than once.  Note that if the --dbdump-include parameter is specified, "
                            "then this parameter will be ignored.")

        parser.add_argument('--dbdump-output', default='/root/data', metavar='PATH',
                            help="Location of output folder for db dumps.  Default is /root/data")

        parser.add_argument('--dump-tables', action='store_true',
                            help="Dump each table for specified databases, in addition to the database "
                            "as a whole.")
        parser.add_argument('--no-dump-tables', action='store_true',
                            help="Do *not* dump each table for each  database.")

        parser.add_argument('--rsync', action='store_true',
                            help="Push all files to remote location via rsync.")
        parser.add_argument('--no-rsync', action='store_true',
                            help="Do NOT push files to remote location via rsync.")

        parser.add_argument('--rsync-include', action='append', dest='rsync_included_prefixes', metavar='PREFIX',
                            help="File prefix which should be included in the rsync run.  Actually, a "
                            "separate rsync call is made for each prefix.  May be specified more than "
                            "once.  Note that if this parameter is used, then ONLY those prefixes specified "
                            "will be rsync'ed.  Also, the --rsync-exclude parameter(s) will still apply.")
        parser.add_argument('--rsync-exclude', action='append', dest='rsync_excluded_prefixes', metavar='PREFIX',
                            help="File prefix which should be excluded from the rsync run.  May be "
                            "specified more than once.  Note that if the --rsync-include parameter is "
                            "specified, then this parameter(s) will still apply also.")

        parser.add_argument('--rsync-remote-host', metavar='ADDRESS',
                            help="Assuming the rsync destination is another server, this should be the "
                            "hostname or IP address of that server.  Or you could set it to an empty string, "
                            "which signifies that the rsync remote destination is another folder on localhost.")
        parser.add_argument('--rsync-remote-prefix', metavar='PREFIX',
                            help="File prefix for the rsync destination.  If --rsync-remote-host is an empty "
                            "string, this prefix must exist on localhost.  Otherwise the prefix must exist on "
                            "the rsync remote host.")

        parser.add_argument('--borg', action='store_true',
                            help="Create backup archive(s) via Borg.")
        parser.add_argument('--no-borg', action='store_true',
                            help="Do NOT create backup archive(s) via Borg.")

        parser.add_argument('--borg-remotes', metavar='REMOTES',
                            help="List of named \"remotes\" for Borg, where archives should be created.")

        parser.add_argument('--borg-include', action='append', dest='borg_included_prefixes', metavar='PREFIX',
                            help="File prefix which should be included in the borg run.  May be specified more "
                            "than once.  If this is *not* specified then --rsync-include will be used, if it is "
                            "specified.  Note that if this parameter is used, then ONLY those prefixes specified "
                            "will be included.  Also, the --borg-exclude parameter(s) will still apply.")
        parser.add_argument('--borg-exclude', action='append', dest='borg_excluded_prefixes', metavar='PREFIX',
                            help="File prefix which should be excluded from the borg run.  May be specified more "
                            "than once.  If this is *not* specified then --rsync-exclude will be used, if it is "
                            "specified.  Note that if the --borg-include parameter is specified, then this "
                            "parameter(s) will still apply also.")

        parser.add_argument('--borg-tag', metavar='TAG',
                            help="Extra \"tag\" to embed within the name of the "
                            "Borg archive which is created.  Note that an archive "
                            "which is \"tagged\" in this way will never be pruned "
                            "by this `rattail backup` command!")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the motions as much as possible, to get an idea "
                            "of what the full backup would do, but don't actually do anything.")

    def run(self, args):

        if args.list_dbs:

            mysql_names = self.get_mysql_names()
            if mysql_names:
                width = max([len(name) for name in mysql_names])
                self.stdout.write("\nmysql\n")
                self.stdout.write("{}\n".format('-' * width))
                for name in mysql_names:
                    self.stdout.write("{}\n".format(name))

            postgres_names = self.get_postgres_names()
            if postgres_names:
                width = max([len(name) for name in postgres_names])
                self.stdout.write("\npostgres\n")
                self.stdout.write("{}\n".format('-' * width))
                for name in postgres_names:
                    self.stdout.write("{}\n".format(name))

            if mysql_names or postgres_names:
                self.stdout.write("\n")
            else:
                self.stdout.write("no databases found\n")

        else: # not listing dbs, so will do dbdump and/or rsync and/or borg

            if args.dbdump and args.no_dbdump:
                raise RuntimeError("Must specify either --dbdump or --no-dbdump, but not both")
            if args.dump_tables and args.no_dump_tables:
                raise RuntimeError("May specify either --dump-tables or --no-dump-tables, but not both")
            if args.rsync and args.no_rsync:
                raise RuntimeError("May specify either --rsync or --no-rsync, but not both")
            if args.borg and args.no_borg:
                raise RuntimeError("May specify either --borg or --no-borg, but not both")

            # we dump dbs by default, unless user or config file says not to
            if not args.no_dbdump and self.config.getbool('rattail.backup', 'dbdump', usedb=False, default=True):

                outdir = args.dbdump_output
                if not outdir:
                    outdir = self.config.get('rattail.backup', 'dbdump.output', default='/root/data')
                if not os.path.exists(outdir):
                    os.mkdir(outdir)

                if args.included_dbs:
                    include = args.included_dbs
                else:
                    include = self.config.getlist('rattail.backup', 'dbdump.include', usedb=False)

                if args.excluded_dbs:
                    exclude = args.excluded_dbs
                else:
                    exclude = self.config.getlist('rattail.backup', 'dbdump.exclude', usedb=False)

                if args.dump_tables:
                    tables = True
                elif args.no_dump_tables:
                    tables = False
                else:
                    tables = self.config.getbool('rattail.backups', 'dbdump.dump_tables', usedb=False, default=False)

                self.dump_mysql_databases(outdir, include=include, exclude=exclude, tables=tables,
                                          dry_run=args.dry_run)
                self.dump_postgres_databases(outdir, include=include, exclude=exclude, tables=tables,
                                             dry_run=args.dry_run)

            # we do *not* run rsync by default, unless user or config file says to
            rsync = False
            rsync_last = False
            if args.rsync or (not args.no_rsync and self.config.getbool('rattail.backup', 'rsync', usedb=False, default=False)):
                rsync = True
                if self.config.getbool('rattail.backup', 'rsync.last', usedb=False, default=False):
                    rsync_last = True

            # we do *not* run borg by default, unless user or config file says to
            borg = False
            if args.borg or (not args.no_borg and self.config.getbool('rattail.backup', 'borg.enabled', usedb=False, default=False)):
                borg = True

            # okay now run rsync and/or borg (in whichever order) depending on params
            if rsync and not rsync_last:
                self.run_rsync(args)
            if borg:
                self.run_borg(args)
            if rsync and rsync_last:
                self.run_rsync(args)

    def run_rsync(self, args):
        self.rsync_files(include=args.rsync_included_prefixes,
                         exclude=args.rsync_excluded_prefixes,
                         dry_run=args.dry_run)

    def run_borg(self, args):
        remotes = (parse_list(args.borg_remotes) or
                   self.config.getlist('rattail.backup', 'borg.remotes',
                                       usedb=False))
        self.make_borg_archives(remotes,
                                include=args.borg_included_prefixes or args.rsync_included_prefixes,
                                exclude=args.borg_excluded_prefixes or args.rsync_excluded_prefixes,
                                tag=args.borg_tag,
                                dry_run=args.dry_run)

    def get_path_includes(self, typ='rsync'):
        return self.config.getlist('rattail.backup', '{}.include'.format(typ), usedb=False, default=[
            '/etc',
            '/home',
            '/opt',
            '/root',
            '/srv',
            '/usr/local',
            '/var',
        ])

    def get_default_excludes(self):
        return [
            '/var/cache/',
        ]

    def get_path_excludes(self, typ='rsync'):
        return self.config.getlist('rattail.backup', '{}.exclude'.format(typ), usedb=False)

    def validate_prefixes(self, prefixes):
        for prefix in prefixes:
            if not prefix.startswith('/'):
                raise RuntimeError("Prefix is not absolute path: {}".format(prefix))

        for prefix in prefixes:
            others = list(prefixes)
            others.remove(prefix)
            for other in others:
                if other.startswith(prefix):
                    raise RuntimeError("Prefix {} is redundant due to prefix {}".format(
                        other, prefix))

    def rsync_files(self, include=None, exclude=None, dry_run=False):
        remote_host = self.config.require('rattail.backup', 'rsync.remote_host', usedb=False)
        remote_prefix = self.config.require('rattail.backup', 'rsync.remote_prefix', usedb=False)
        if not remote_prefix.startswith('/'):
            raise RuntimeError("Remote prefix is not absolute path: {}".format(remote_prefix))
        if remote_host:
            remote_prefix = '{}:{}'.format(remote_host, remote_prefix)

        if include:
            prefixes = include
        else:
            prefixes = self.get_path_includes(typ='rsync')
        self.validate_prefixes(prefixes)

        if exclude is None:
            exclude = self.get_path_excludes(typ='rsync')
            if not exclude:
                exclude = self.get_default_excludes()

        for prefix in prefixes:
            if not os.path.exists(prefix):
                log.warning("skipping prefix which doesn't exist locally: %s", prefix)
                continue

            excluded = []
            if exclude:
                excluded = [pattern for pattern in exclude if pattern.startswith(prefix)]
            log.info("running rsync for prefix: %s", prefix)

            parent = os.path.dirname(prefix)
            if parent == '/':
                destination = remote_prefix
                excludes = ['--exclude={}'.format(pattern) for pattern in excluded]

            else: # prefix parent is not root (/) dir
                destination = '{}{}/'.format(remote_prefix, parent)

                # exclusion patterns must apparently be relative in this case,
                # i.e. to the topmost folder being synced
                excludes = []
                for pattern in excluded:
                    pattern = '/'.join(pattern.split('/')[2:])
                    excludes.append('--exclude={}'.format(pattern))

                # and must also create the parent folder on remote side
                cmd = ['rsync', '--dirs', '--perms', '--times', '--group', '--owner', parent, remote_prefix]
                log.debug("rsync command is: %s", cmd)
                if not dry_run:
                    subprocess.check_call(cmd)

            # okay let's rsync
            rsync = ['rsync', '--archive', '--delete-during', '--partial']
            if self.verbose:
                rsync.append('--progress')
            cmd = rsync + excludes + [prefix, destination]
            log.debug("rsync command is: %s", cmd)
            if not dry_run:
                try:
                    subprocess.check_call(cmd)
                except subprocess.CalledProcessError as error:
                    # nb. rsync exits with code 24 when scenario is
                    # "Partial transfer due to vanished source files"
                    # but that just means all that could be synced, was
                    # synced, though some could not be synced.  we do
                    # not consider that a serious issue, so want to
                    # avoid the outright error here.
                    if error.returncode != 24:
                        raise
                    log.warning("rsync command exited with code 24")

    def make_borg_archives(self, remotes, include=None, exclude=None,
                           tag=None, dry_run=False):
        borg = self.config.get('rattail.backup', 'borg', usedb=False)
        if not borg:
            borg = os.path.join(os.path.dirname(sys.executable), 'borg')
            if not os.path.exists(borg):
                borg = subprocess.check_output('which borg')

        if include:
            prefixes = include
        else:
            prefixes = self.get_path_includes(typ='borg')
            if not prefixes:
                prefixes = self.get_path_includes(typ='rsync')
        self.validate_prefixes(prefixes)

        if exclude is None:
            exclude = self.get_path_excludes(typ='borg')
            if not exclude:
                exclude = self.get_path_excludes(typ='rsync')
                if not exclude:
                    exclude = self.get_default_excludes()

        hostname = self.config.get('rattail.backup', 'borg.archive_hostname', usedb=False)
        if not hostname:
            hostname = socket.gethostname()

        for remote in remotes:

            # borg_env
            repo = self.config.require('rattail.backup', 'borg.remote.{}.repo'.format(remote), usedb=False)
            passphrase = self.config.require('rattail.backup', 'borg.remote.{}.passphrase'.format(remote), usedb=False)
            remote_borg = self.config.get('rattail.backup', 'borg.remote.{}.borg'.format(remote), usedb=False)
            borg_env = {
                'HOME': os.environ['HOME'],
                'BORG_REPO': repo,
                'BORG_PASSPHRASE': passphrase,
                'BORG_REMOTE_PATH': remote_borg or 'borg',
            }

            # borg create
            cmd = [borg, 'create',
                   '--compression', 'lz4',
                   '--exclude-caches']
            if self.verbose:
                cmd.extend([
                    '--verbose',
                    '--list',
                    '--filter', 'AME',
                    '--stats',
                    '--show-rc',
                ])

            for pattern in exclude:
                cmd.extend(['--exclude', pattern])

            if tag:
                # note that in this case our hostname will be immediately
                # followed by a colon instead of hyphen.  this is what "breaks"
                # the typical pattern and ensures this tagged archive will not
                # be pruned under typical command runs.
                archive_name = '::{}:{}-{{now}}'.format(hostname, tag)
            else:
                archive_name = '::{}-{{now}}'.format(hostname)
            cmd.append(archive_name)

            for prefix in prefixes:
                cmd.append(prefix)

            # run `borg create`
            log.debug("borg create command is: %s", cmd)
            if not dry_run:
                try:
                    subprocess.check_call(cmd, env=borg_env)
                except subprocess.CalledProcessError as error:
                    # nb. borg exits with code 1 for "warning" but
                    # this is different than "error" - cf. also
                    # https://borgbackup.readthedocs.io/en/latest/man_intro.html#return-codes
                    if error.returncode != 1:
                        raise
                    log.warning("borg create command exited with code 1")

            # the only thing left is to prune, which if we're tagging, we
            # probably should just skip.  maybe let command line override?
            if tag:
                return

            # generate --keep args for prune, based on config
            keep = []
            keep_defaults = {
                'daily': '7',
                'weekly': '4',
                'monthly': '6',
            }
            for typ, default in keep_defaults.items():
                value = self.config.get('rattail.backup', 'borg.keep.{}'.format(typ))
                if value is None:
                    keep.extend(['--keep-{}'.format(typ), default])
                elif value.isdigit():
                    keep.extend(['--keep-{}'.format(typ), int(value)])
                elif parse_bool(value):
                    keep.extend(['--keep-{}'.format(typ), default])

            # borg prune
            cmd = [borg, 'prune',
                   '--glob-archives', '{}-*'.format(hostname)] + keep
            if self.verbose:
                cmd.extend([
                    '--list',
                    '--show-rc',
                ])

            log.debug("borg prune command is: %s", cmd)
            if not dry_run:
                subprocess.check_call(cmd, env=borg_env)

    def get_mysql_names(self, include=None, exclude=None, exclude_builtin=True):
        try:
            output = subprocess.check_output([
                'mysql', '--execute', "show databases;",
                '--batch', '--skip-column-names',
            ])
        except Exception as error:
            if isinstance(error, FileNotFoundError):
                # assuming there is no mysql binary
                return []
            raise
        output = output.decode('ascii') # TODO: how to detect this etc.?

        names = output.split('\n')
        names = [name.strip() for name in names if name.strip()]

        if exclude_builtin:
            builtins = [
                'mysql',
                'information_schema',
                'performance_schema',
                'sys',
            ]
            for name in builtins:
                if name in names:
                    names.remove(name)

        return self.filter_names(names, include, exclude)

    def get_mysql_tables(self, database):
        try:
            output = subprocess.check_output([
                'mysql', '--execute', "show tables;",
                '--batch', '--skip-column-names',
                database,
            ])
        except Exception as error:
            if isinstance(error, FileNotFoundError):
                # assuming there is no mysql binary
                return []
            raise
        output = output.decode('ascii') # TODO: how to detect this etc.?

        names = output.split('\n')
        names = [name.strip() for name in names if name.strip()]
        return names

    def get_postgres_names(self, include=None, exclude=None, exclude_builtin=True):
        """
        Returns a list of database names found in PostgreSQL, filtered
        according to the various arguments provided.
        """
        names = []

        # can only use `psql` if it's present, so check first
        with open(os.devnull, 'w') as devnull:
            psql_exists = subprocess.call(['which', 'psql'], stdout=devnull) == 0
        if psql_exists:

            output = subprocess.check_output([
                'sudo', '-u', 'postgres', 'psql', '--list', '--tuples-only',
            ])
            output = output.decode('ascii') # TODO: how to detect this etc.?

            for line in output.split('\n'):
                line = line.strip()
                if line and '|' in line and not line.startswith('|'):
                    name = line.split('|')[0].strip()
                    names.append(name)

        if names and exclude_builtin:
            builtins = [
                'postgres',
                'template0',
                'template1',
            ]
            for name in builtins:
                if name in names:
                    names.remove(name)

        return self.filter_names(names, include, exclude)

    def filter_names(self, names, include, exclude):
        if include:
            names = [name for name in names if name in include]
        elif exclude:
            names = [name for name in names if name not in exclude]
        return names

    def dump_mysql_databases(self, outdir, include=None, exclude=None, tables=False, dry_run=False):
        names = self.get_mysql_names(include=include, exclude=exclude)
        for name in names:
            log.info("dumping mysql db: %s", name)

            # generate mysqldump command
            sql_path = os.path.join(outdir, '{}-mysql.sql'.format(name))
            mysqldump = ['mysqldump']
            if self.verbose:
                mysqldump.append('--verbose')
            cmd = mysqldump + [name, '--result-file={}'.format(sql_path)]
            log.debug("mysqldump command is: %s", cmd)

            # run mysqldump command
            if not dry_run:
                subprocess.check_call(cmd)
                subprocess.check_call(['gzip', '--force', sql_path])
                gz_path = "{}.gz".format(sql_path)
                log.debug("database dump completed, file is at %s", gz_path)

            if tables:

                # first, forcibly recreate 'tables' dir for this db
                if not dry_run:
                    tablesdir = os.path.join(outdir, 'tables', 'mysql-{}'.format(name))
                    if os.path.exists(tablesdir):
                        shutil.rmtree(tablesdir)
                    os.makedirs(tablesdir)

                # then make a dump of each table reported by 'show tables'
                for table in self.get_mysql_tables(name):
                    log.info("dumping mysql table: %s", table)

                    # generate mysqldump command
                    sql_path = os.path.join(tablesdir, '{}.sql'.format(table))
                    mysqldump = ['mysqldump']
                    if self.verbose:
                        mysqldump.append('--verbose')
                    cmd = mysqldump + [name, table, '--result-file={}'.format(sql_path)]
                    log.debug("mysqldump command is: %s", cmd)

                    # run mysqldump command
                    if not dry_run:
                        subprocess.check_call(cmd)
                        subprocess.check_call(['gzip', '--force', sql_path])
                        gz_path = "{}.gz".format(sql_path)
                        log.debug("table dump completed, file is at %s", gz_path)

    def dump_postgres_databases(self, outdir, include=None, exclude=None, tables=False, dry_run=False):
        names = self.get_postgres_names(include=include, exclude=exclude)
        for name in names:
            log.info("dumping postgres db: %s", name)
            if not dry_run:
                sql_path = os.path.join(outdir, '{}-postgres.sql'.format(name))
                tmp_path = '/tmp/{}.sql'.format(name)
                if os.path.exists(tmp_path):
                    os.remove(tmp_path)
                pg_dump = ['sudo', '-u', 'postgres', 'pg_dump']
                if self.verbose:
                    pg_dump.append('--verbose')
                cmd = pg_dump + [name, '--file={}'.format(tmp_path)]
                log.debug("pg_dump command is: %s", cmd)
                subprocess.check_call(cmd)
                subprocess.check_call(['chown', 'root:', tmp_path])
                subprocess.check_call(['mv', tmp_path, sql_path])
                subprocess.check_call(['gzip', '--force', sql_path])
                gz_path = "{}.gz".format(sql_path)
                log.debug("dump completed, file is at %s", gz_path)

        if tables:
            log.info("TODO: should honor --dump-tables flag for postgres")



# -*- coding: utf-8 -*-
"""add purchase shipped counts

Revision ID: 977a76f259bd
Revises: 92a49edc45c3
Create Date: 2018-05-16 16:06:13.691091

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '977a76f259bd'
down_revision = u'92a49edc45c3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('cases_shipped', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('units_shipped', sa.Numeric(precision=10, scale=4), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('cases_shipped', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('units_shipped', sa.Numeric(precision=10, scale=4), nullable=True))


def downgrade():

    # purchase_item
    op.drop_column('purchase_item', 'units_shipped')
    op.drop_column('purchase_item', 'cases_shipped')

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'units_shipped')
    op.drop_column('purchase_batch_row', 'cases_shipped')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Mail Monitor Actions
"""

import os
import shutil
import tempfile

from rattail.monitoring import MonitorAction
from rattail.config import parse_bool
from rattail.files import locking_copy


class MessageAction(MonitorAction):
    """
    Base class for mailmon message actions.
    """

    def __call__(self, server, msguid, *args, **kwargs):
        """
        This method must be implemented in the subclass; it defines
        what the action actually *does*.  The monitor daemon will
        invoke this method for all new messages which are discovered.

        :param server: Reference to the ``imaplib.server`` instance,
           which will be connected with an active session.

        :param msguid: UID for the message upon which to act.
        """
        raise NotImplementedError


def download_message(server, msg_uid, output_dir, locking=False):
    """
    Simple action to "download" a message to local filesystem.

    :param output_dir: Path to the folder into which message should be
       written.  Note that the filename will be like
       ``{msg_uid}.eml``.

    :param locking: Flag to indicate that the
       :func:`rattail.files.locking_copy()` function should be used to
       place the file in its final location.  This is useful if you
       then also have a rattail filemon watching the ``output_dir``.
    """
    if not isinstance(locking, bool):
        locking = parse_bool(locking)

    # fetch message data
    code, msg_data = server.uid('fetch', msg_uid, '(RFC822)')
    if code != 'OK':
        raise RuntimeError("IMAP4.fetch() for msg_uid %s returned "
                           "bad code %s - msg_data is: %s",
                           msg_uid, code, msg_data)

    # extract message body
    response, msg_body = msg_data[0]

    # figure out where we need to write the file
    # nb. msg_uid is bytes, must convert
    filename = '{}.eml'.format(msg_uid.decode('utf_8'))
    if locking:
        tempdir = tempfile.mkdtemp()
        path = os.path.join(tempdir, filename)
    else: # no locking, write directly to file
        path = os.path.join(output_dir, filename)

    # write message to file
    with open(path, 'wb') as f:
        f.write(msg_body)

    # maybe move temp file to final path
    if locking:
        locking_copy(path, output_dir)
        shutil.rmtree(tempdir)
    

def move_message(server, msguid, newfolder):
    """
    Simple action to "move" a message to another IMAP folder, on the
    same server.
    """
    # copy msg to new folder
    code, response = server.uid('COPY', msguid, newfolder)
    if code != 'OK':
        raise RuntimeError("IMAP.copy(uid={}) returned bad code: {}".format(
            msguid, code))

    # mark old msg as deleted
    code, response = server.uid('STORE', msguid, '+FLAGS', '(\Deleted)')
    if code != 'OK':
        raise RuntimeError("IMAP.store(uid={}) returned bad code: {}".format(
            msguid, code))

    # expunge deleted messages
    code, response = server.expunge()
    if code != 'OK':
        raise RuntimeError("IMAP.expunge() returned bad code: {}".format(code))



# -*- coding: utf-8 -*-
"""add product.deleted

Revision ID: 1513dc6d6ca7
Revises: 47d30c955111
Create Date: 2015-02-10 22:17:45.367561

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '1513dc6d6ca7'
down_revision = u'47d30c955111'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # add product.deleted
    op.add_column('product', sa.Column('deleted', sa.Boolean(), nullable=True))
    product = sa.sql.table('product', sa.sql.column('deleted'))
    op.execute(product.update().values({'deleted': False}))
    op.alter_column('product', 'deleted', nullable=False,
                    existing_type=sa.Boolean())
    op.add_column('product_version', sa.Column('deleted', sa.Boolean(), nullable=True))


def downgrade():

    # drop product.deleted
    op.drop_column('product_version', 'deleted')
    op.drop_column('product', 'deleted')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Types
"""

import json

from sqlalchemy import types

from rattail.gpc import GPC


class GPCType(types.TypeDecorator):
    """
    SQLAlchemy type engine for GPC data.
    """
    impl = types.BigInteger
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return int(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return GPC(value)


class JSONTextDict(types.TypeDecorator):
    """
    SQLAlchemy type engine for JSON data.  Interprets "raw" text as JSON when
    reading from DB, and writes JSON text back to DB.

    Note that the Python value for a field with this type will appear as a dict.
    """
    impl = types.Text
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Project Commands
"""

from __future__ import unicode_literals, absolute_import

import os
import subprocess
import sys

from rattail.commands.core import Subcommand
from rattail.projects.handler import RattailProjectHandler


class MakeProject(Subcommand):
    """
    Make a new app project
    """
    name = 'make-project'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        parser.add_argument('path',
                            help="Path to project folder")

    def run(self, args):

        path = os.path.abspath(args.path)
        if os.path.exists(path):
            self.stderr.write("path already exists: {}\n".format(path))
            sys.exit(1)

        # nb. technically these would be required and auto-installed
        # as needed (later), but seems better to do this explicitly
        # up-front before any other command output
        self.require_prompt_toolkit()
        self.require_rich()

        # welcome, continue?
        self.rprint("\n\t[blue]Welcome to Rattail[/blue]")
        self.rprint("\n\tThis tool will generate source code for a new project.")
        if not self.basic_prompt("continue?", True, is_bool=True):
            self.rprint()
            sys.exit(0)

        # name
        name = os.path.basename(path)

        # app_table_prefix
        prefix = name
        prefix = prefix.replace('-', '_')
        prefix = prefix.replace(' ', '_')
        app_table_prefix = self.basic_prompt('app_table_prefix',
                                             default=prefix)
        app_table_prefix = app_table_prefix.rstrip('_')

        # app_class_prefix
        prefix = name
        prefix = prefix.replace('-', '_')
        prefix = prefix.replace('_', ' ')
        app_class_prefix = ''.join([w.capitalize() for w in prefix.split()])
        app_class_prefix = self.basic_prompt('app_class_prefix',
                                             default=app_class_prefix)

        # org_name
        org_name = self.basic_prompt('org_name', required=True)

        # pypi_name
        pypi_name = name
        pypi_name = pypi_name.replace('_', ' ')
        pypi_name = pypi_name.replace('-', ' ')
        pypi_name = '-'.join([w.capitalize()
                              for w in org_name.split() + pypi_name.split()])
        pypi_name = self.basic_prompt('pypi_name', default=pypi_name)

        # app_title
        app_title = name
        app_title = app_title.replace('-', ' ')
        app_title = app_title.replace('_', ' ')
        app_title = ' '.join([w.capitalize() for w in app_title.split()])

        # generate project
        project_handler = RattailProjectHandler(self.config)
        options = {
            'name': app_title,
            'slug': name,
            'organization': org_name,
            'python_project_name': pypi_name,
            'python_name': app_table_prefix,
            'app_class_prefix': app_class_prefix,
            'has_db': True,
            'extends_db': True,
            'has_web': True,

            # TODO: these should not be needed..?
            'has_web_api': False,
            'has_datasync': False,
            'integrates_catapult': False,
            'integrates_corepos': False,
            'integrates_locsms': False,
            'uses_fabric': False,
        }
        project_handler.generate_project('rattail', name, options, path=path)
        self.rprint("\n\tproject created at:  [bold green]{}[/bold green]".format(
            path))

        # install pkg
        if self.basic_prompt("install project package?", is_bool=True, default=True):
            subprocess.check_call(['pip', 'install', '-e', path])
            self.rprint("\n\tpackage installed:  [bold green]{}[/bold green]".format(
                pypi_name))

            self.rprint("\n\tinstall and configure the app with:")
            self.rprint("\n\t[blue]{} -n install[/blue]".format(name))

        self.rprint()



# -*- coding: utf-8 -*-
"""add Person.middle_name

Revision ID: 0b90d5c59fd7
Revises: 3df3144cec58
Create Date: 2016-01-07 18:51:40.499882

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '0b90d5c59fd7'
down_revision = u'3df3144cec58'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # person
    op.add_column('person', sa.Column('middle_name', sa.String(length=50), nullable=True))


def downgrade():

    # person
    op.drop_column('person', 'middle_name')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor invoice parser for United Natural Foods (UNFI)
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six

from rattail.db import model
from rattail.gpc import GPC
from rattail.vendors.invoices import InvoiceParser
from rattail.csvutil import UnicodeDictReader, UnicodeReader


class UnfiInvoiceParser(InvoiceParser):
    """
    Parser for UNFI CSV invoice files.
    """
    key = 'rattail.contrib.unfi'
    display = "United Natural Foods (UNFI)"
    vendor_key = 'unfi'

    def parse_invoice_date(self, data_path):
        if six.PY3:
            csv_file = open(data_path, 'rt', encoding='latin_1')
            reader = UnicodeDictReader(csv_file)
        else: # PY2
            csv_file = open(data_path, 'rb')
            reader = UnicodeDictReader(csv_file, encoding='latin_1')
        data = next(reader)
        csv_file.close()
        return datetime.datetime.strptime(data['InvoiceDate'], '%m/%d/%Y').date()

    def parse_rows(self, data_path):
        if six.PY3:
            csv_file = open(data_path, 'rt', encoding='latin_1')
        else: # PY2
            csv_file = open(data_path, 'rb')

        # TODO: The following logic is largely copied from the Scan Genius
        # order parser (in `rattail_livnat.scangenius`).  I wonder if we
        # can abstract it into a generic yet custom CSV parser...?

        # We want to ignore the header section here.  However the use of
        # UnicodeDictReader below requires iteration, and yet also expects
        # the "beginning" of the file to contain the fieldnames.  To force
        # this all to play nicely, we calculate our offset and then seek to
        # it explicitly.  It's probably likely that there is a better way,
        # but this works.
        offset = self.find_details_offset(csv_file)
        csv_file.seek(offset)

        if six.PY3:
            reader = UnicodeDictReader(csv_file)
        else: # PY2
            reader = UnicodeDictReader(csv_file, encoding='latin_1')
        for data in reader:

            # Only consider 'Detail' rows; this check is mostly for the
            # sake of ignoring 'Footer' rows.
            if data['RecType'] != 'Detail':
                continue

            row = model.VendorInvoiceBatchRow()
            row.item_entry = data['UPC'].replace('-', '')
            row.upc = GPC(row.item_entry)
            row.vendor_code = data['ProductID']
            row.brand_name = data['Brand']
            row.description = data['Description']
            row.size = data['Size']
            row.case_quantity = self.int_(data['Servings'])
            row.ordered_cases = int(self.decimal(data['QuantityOrdered']))
            row.shipped_cases = self.int_(data['QuantityShipped'])
            row.out_of_stock = True if data['Status'] == 'OUT' else False
            row.unit_cost = self.decimal(data['NetPricePerUnit'])
            row.total_cost = self.decimal(data['ExtendedPrice'])
            row.case_cost = row.unit_cost * row.case_quantity

            yield row

        csv_file.close()

    def find_details_offset(self, csv_file):
        """
        Find the character offset of the details data within a CSV file.
        """
        # TODO: The following logic is largely copied from the Scan Genius
        # order parser (in ``rattail_livnat.scangenius``).  I wonder if we can
        # abstract it into a generic yet custom CSV parser...?

        offset = 0
        for line, row in enumerate(csv_file, 1):
            offset += len(row)
            if line == 2:
                break

        # account for newline chars
        offset += 2 * (len(csv_file.newlines) - 1)

        return offset



# -*- coding: utf-8 -*-
"""add role.adminish

Revision ID: d39ff41238fa
Revises: 363a96293285
Create Date: 2022-08-21 19:29:09.547336

"""

# revision identifiers, used by Alembic.
revision = 'd39ff41238fa'
down_revision = '363a96293285'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # role
    op.add_column('role', sa.Column('adminish', sa.Boolean(), nullable=True))
    op.add_column('role_version', sa.Column('adminish', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # role
    op.drop_column('role_version', 'adminish')
    op.drop_column('role', 'adminish')



# -*- coding: utf-8 -*-
"""add user.last_login

Revision ID: 37e0e46ec6fe
Revises: 23600f00cda7
Create Date: 2017-10-23 14:55:50.162386

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '37e0e46ec6fe'
down_revision = u'23600f00cda7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # user
    op.add_column('user', sa.Column('last_login', sa.DateTime(), nullable=True))


def downgrade():

    # user
    op.drop_column('user', 'last_login')



# -*- coding: utf-8; -*-
"""add ifps_plu

Revision ID: ef398701ffc6
Revises: c8981df473d0
Create Date: 2020-12-06 16:54:34.622596

"""

# revision identifiers, used by Alembic.
revision = 'ef398701ffc6'
down_revision = 'c8981df473d0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # ifps_plu
    op.create_table('ifps_plu',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('plu', sa.Integer(), nullable=False),
                    sa.Column('category', sa.String(length=100), nullable=True),
                    sa.Column('commodity', sa.String(length=100), nullable=True),
                    sa.Column('variety', sa.Text(), nullable=True),
                    sa.Column('size', sa.String(length=100), nullable=True),
                    sa.Column('measurements_north_america', sa.String(length=255), nullable=True),
                    sa.Column('measurements_rest_of_world', sa.String(length=255), nullable=True),
                    sa.Column('restrictions_notes', sa.Text(), nullable=True),
                    sa.Column('botanical_name', sa.String(length=100), nullable=True),
                    sa.Column('aka', sa.Text(), nullable=True),
                    sa.Column('revision_date', sa.DateTime(), nullable=True),
                    sa.Column('date_added', sa.DateTime(), nullable=True),
                    sa.Column('gpc', sa.String(length=100), nullable=True),
                    sa.Column('image', sa.String(length=255), nullable=True),
                    sa.Column('image_source', sa.String(length=100), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('plu', name='ifps_plu_uq_plu')
    )
    op.create_table('ifps_plu_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('plu', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('category', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('commodity', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('variety', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('size', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('measurements_north_america', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('measurements_rest_of_world', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('restrictions_notes', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('botanical_name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('aka', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('revision_date', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('date_added', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('gpc', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('image', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('image_source', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_ifps_plu_version_end_transaction_id'), 'ifps_plu_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_ifps_plu_version_operation_type'), 'ifps_plu_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_ifps_plu_version_transaction_id'), 'ifps_plu_version', ['transaction_id'], unique=False)


def downgrade():

    # ifps_plu
    op.drop_index(op.f('ix_ifps_plu_version_transaction_id'), table_name='ifps_plu_version')
    op.drop_index(op.f('ix_ifps_plu_version_operation_type'), table_name='ifps_plu_version')
    op.drop_index(op.f('ix_ifps_plu_version_end_transaction_id'), table_name='ifps_plu_version')
    op.drop_table('ifps_plu_version')
    op.drop_table('ifps_plu')



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

import six
import sqlalchemy as sa

from rattail import products as mod
from rattail.config import make_config
from rattail.gpc import GPC
from rattail.db import Session


class TestProductsHandler(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.handler = self.make_handler()

    def make_config(self):
        return make_config([], extend=False)

    def make_handler(self):
        return mod.ProductsHandler(self.config)

    def test_make_gpc(self):

        # basic real-world example
        result = self.handler.make_gpc('074305001321')
        self.assertIsInstance(result, GPC)
        self.assertEqual(six.text_type(result), '00074305001321')

        # and let it calculate check digit
        result = self.handler.make_gpc('7430500132', calc_check_digit='upc')
        self.assertIsInstance(result, GPC)
        self.assertEqual(six.text_type(result), '00074305001321')

        # can also pass integer, and let it auto-calc check digit
        # (b/c data length makes it clear check digit is missing)
        result = self.handler.make_gpc(7430500132)
        self.assertIsInstance(result, GPC)
        self.assertEqual(six.text_type(result), '00074305001321')

        # bad one should raise error
        self.assertRaises(ValueError, self.handler.make_gpc, 'BAD_VALUE')

        # unless we suppress errors
        result = self.handler.make_gpc('BAD_VALUE', ignore_errors=True)
        self.assertIsNone(result)

    def test_make_full_description(self):
        model = self.config.get_model()

        # basic example
        result = self.handler.make_full_description(
            brand_name="Bragg's",
            description="Apple Cider Vinegar",
            size="32oz")
        self.assertEqual(result, "Bragg's Apple Cider Vinegar 32oz")

        # product example
        product = model.Product(
            brand=model.Brand(name="Bragg's"),
            description="Apple Cider Vinegar",
            size="32oz")
        result = self.handler.make_full_description(product)
        self.assertEqual(result, "Bragg's Apple Cider Vinegar 32oz")

        # pending product example
        product = model.PendingProduct(
            brand_name="Bragg's",
            description="Apple Cider Vinegar",
            size="32oz")
        result = self.handler.make_full_description(product)
        self.assertEqual(result, "Bragg's Apple Cider Vinegar 32oz")

    def test_get_case_size(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)

        # no case size by default
        product = model.Product()
        result = self.handler.get_case_size(product)
        self.assertIsNone(result)

        # main attr example
        product = model.Product(case_size=12)
        result = self.handler.get_case_size(product)
        self.assertEqual(result, 12)

        # cost attr example
        product = model.Product()
        vendor = model.Vendor()
        product.costs.append(model.ProductCost(vendor=vendor,
                                               case_size=24))
        session.add(product)
        session.flush()
        result = self.handler.get_case_size(product)
        self.assertEqual(result, 24)

    def test_get_url(self):
        app = self.config.get_app()
        model = self.config.get_model()

        # no url by default
        product = model.Product()
        result = self.handler.get_url(product)
        self.assertIsNone(result)

        # basic example
        self.config.setdefault('rattail', 'base_url', 'http://example.com')
        uuid = app.make_uuid()
        product = model.Product(uuid=uuid)
        result = self.handler.get_url(product)
        self.assertEqual(result, 'http://example.com/products/{}'.format(uuid))

    def test_locate_product_for_alt_code(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # setup data
        product1 = model.Product()
        product1.codes.append('12345')
        session.add(product1)
        product2 = model.Product()
        product2.codes.append('67890')
        session.add(product2)
        session.flush()

        # unknown code returns nothing
        result = self.handler.locate_product_for_alt_code(session, 'bogus')
        self.assertIsNone(result)

        # basic example
        result = self.handler.locate_product_for_alt_code(session, '12345')
        self.assertIs(result, product1)

    def test_locate_product_for_vendor_code(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # setup data
        vendor = model.Vendor()
        product1 = model.Product()
        product1.costs.append(model.ProductCost(vendor=vendor, code='12345'))
        session.add(product1)
        product2 = model.Product()
        product2.costs.append(model.ProductCost(vendor=vendor, code='67890'))
        session.add(product2)
        session.flush()

        # unknown code returns nothing
        result = self.handler.locate_product_for_vendor_code(session, 'bogus')
        self.assertIsNone(result)

        # basic example, w/ vendor restriction
        result = self.handler.locate_product_for_vendor_code(session, '12345',
                                                             vendor=vendor)
        self.assertIs(result, product1)

        # basic example, for any vendor
        result = self.handler.locate_product_for_vendor_code(session, '67890')
        self.assertIs(result, product2)

    def test_locate_product_for_gpc(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # setup data
        product1 = model.Product(upc=app.make_gpc('00074305001321'))
        session.add(product1)
        product2 = model.Product(upc=app.make_gpc('0021234500000',
                                                  calc_check_digit='upc'))
        session.add(product2)
        session.flush()

        # return nothing if no gpc provided
        result = self.handler.locate_product_for_gpc(session, None)
        self.assertIsNone(result)

        # basic example
        result = self.handler.locate_product_for_gpc(
            session, app.make_gpc('00074305001321'))
        self.assertIs(result, product1)

        # type2 lookup does not happen by default
        type2 = app.make_gpc('0021234501299', calc_check_digit='upc')
        result = self.handler.locate_product_for_gpc(session, type2)
        self.assertIsNone(result)

        # but we can enable type2 lookup, then it should work
        self.config.setdefault('rattail', 'products.convert_type2_for_gpc_lookup',
                               'true')
        result = self.handler.locate_product_for_gpc(session, type2)
        self.assertIs(result, product2)

    def test_locate_product_for_upc(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # setup data
        product1 = model.Product(upc=app.make_gpc('00074305001321'))
        session.add(product1)
        product2 = model.Product(upc=app.make_gpc('0021234500000',
                                                  calc_check_digit='upc'))
        session.add(product2)
        product3 = model.Product(upc=app.make_gpc('00035200007122'))
        session.add(product3)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_upc(session, None)
        self.assertIsNone(result)

        # bad data returns nothing
        result = self.handler.locate_product_for_upc(session, 'bogus')
        self.assertIsNone(result)

        # basic UPC-A, with check digit
        result = self.handler.locate_product_for_upc(session, '074305001321')
        self.assertIs(result, product1)

        # UPC-A without check digit
        result = self.handler.locate_product_for_upc(session, '07430500132')
        self.assertIs(result, product1)

        # UPC-E
        result = self.handler.locate_product_for_upc(session, '03571222')
        self.assertIs(result, product3)

        # basic GPC
        result = self.handler.locate_product_for_upc(
            session, app.make_gpc('00074305001321'))
        self.assertIs(result, product1)

        # type2 GPC lookup does not happen by default
        type2 = app.make_gpc('0021234501299', calc_check_digit='upc')
        result = self.handler.locate_product_for_upc(session, type2)
        self.assertIsNone(result)

        # but we can enable type2 lookup, then it should work
        self.config.setdefault('rattail', 'products.convert_type2_for_gpc_lookup',
                               'true')
        result = self.handler.locate_product_for_upc(session, type2)
        self.assertIs(result, product2)

    def test_locate_product_for_item_id(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # setup data
        product1 = model.Product(item_id='0007430500132')
        session.add(product1)
        product2 = model.Product(item_id='0021234500000')
        session.add(product2)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_item_id(session, None)
        self.assertIsNone(result)

        # basic example
        result = self.handler.locate_product_for_item_id(session, '0007430500132')
        self.assertIs(result, product1)

    def test_locate_product_for_scancode(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # setup data
        product1 = model.Product(scancode='074305001321')
        session.add(product1)
        product2 = model.Product(scancode='074305001161')
        session.add(product2)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_scancode(session, None)
        self.assertIsNone(result)

        # basic example
        result = self.handler.locate_product_for_scancode(session, '074305001321')
        self.assertIs(result, product1)

    def test_locate_product_for_key_upc(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # set upc as product key
        self.config.setdefault('rattail', 'product.key', 'upc')

        # setup data
        product1 = model.Product(upc=app.make_gpc('00074305001321'))
        session.add(product1)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_key(session, None)
        self.assertIsNone(result)

        # basic UPC-A, with check digit
        result = self.handler.locate_product_for_key(session, '074305001321')
        self.assertIs(result, product1)

        # UPC-A without check digit
        result = self.handler.locate_product_for_key(session, '07430500132')
        self.assertIs(result, product1)

    def test_locate_product_for_key_item_id(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # set item_id as product key
        self.config.setdefault('rattail', 'product.key', 'item_id')

        # setup data
        product1 = model.Product(item_id='0007430500132')
        session.add(product1)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_key(session, None)
        self.assertIsNone(result)

        # basic example
        result = self.handler.locate_product_for_key(session, '0007430500132')
        self.assertIs(result, product1)

    def test_locate_product_for_key_scancode(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # set scancode as product key
        self.config.setdefault('rattail', 'product.key', 'scancode')

        # setup data
        product1 = model.Product(scancode='074305001321')
        session.add(product1)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_key(session, None)
        self.assertIsNone(result)

        # basic example
        result = self.handler.locate_product_for_key(session, '074305001321')
        self.assertIs(result, product1)

    def test_locate_product_for_entry(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # set upc as product key
        self.config.setdefault('rattail', 'product.key', 'upc')

        # setup data
        vendor = model.Vendor()
        product1 = model.Product(upc=app.make_gpc('00074305001321'))
        session.add(product1)
        product2 = model.Product(upc=app.make_gpc('0021234500000',
                                                  calc_check_digit='upc'))
        session.add(product2)
        product3 = model.Product(upc=app.make_gpc('00035200007122'))
        session.add(product3)
        product4 = model.Product(item_id='42')
        product4.codes.append('4242')
        session.add(product4)
        product5 = model.Product(scancode='074305001161')
        product5.costs.append(model.ProductCost(vendor=vendor, code='12345'))
        session.add(product5)
        session.flush()

        # ask for nothing, get nothing
        result = self.handler.locate_product_for_entry(session, None)
        self.assertIsNone(result)

        # bad data returns nothing
        result = self.handler.locate_product_for_entry(session, 'bogus',
                                                       lookup_fields=[
                                                           'uuid',
                                                           'upc',
                                                           'item_id',
                                                           'scancode',
                                                           'vendor_code',
                                                           'alt_code',
                                                       ])
        self.assertIsNone(result)

        # basic UUID test
        result = self.handler.locate_product_for_entry(session, product5.uuid)
        self.assertIs(result, product5)

        # basic UPC-A, with check digit
        result = self.handler.locate_product_for_entry(session, '074305001321')
        self.assertIs(result, product1)

        # UPC-A without check digit
        result = self.handler.locate_product_for_entry(session, '07430500132')
        self.assertIs(result, product1)

        # UPC-E
        result = self.handler.locate_product_for_entry(session, '03571222')
        self.assertIs(result, product3)

        # basic GPC
        result = self.handler.locate_product_for_entry(
            session, app.make_gpc('00074305001321'))
        self.assertIs(result, product1)

        # type2 GPC lookup does not happen by default
        type2 = app.make_gpc('0021234501299', calc_check_digit='upc')
        result = self.handler.locate_product_for_entry(session, type2)
        self.assertIsNone(result)

        # but we can enable type2 lookup, then it should work
        self.config.setdefault('rattail', 'products.convert_type2_for_gpc_lookup',
                               'true')
        result = self.handler.locate_product_for_entry(session, type2)
        self.assertIs(result, product2)

        # basic item_id test
        result = self.handler.locate_product_for_entry(session, '42',
                                                       lookup_fields=['item_id'])
        self.assertIs(result, product4)

        # basic scancode test
        result = self.handler.locate_product_for_entry(session, '074305001161',
                                                       lookup_fields=['scancode'])
        self.assertIs(result, product5)

        # basic alt code test
        result = self.handler.locate_product_for_entry(session, '4242',
                                                       lookup_fields=['alt_code'])
        self.assertIs(result, product4)

        # basic vendor code test, w/ vendor restriction
        result = self.handler.locate_product_for_entry(session, '12345',
                                                       lookup_fields=['vendor_code'],
                                                       vendor=vendor)
        self.assertIs(result, product5)

        # basic vendor code test, any vendor
        result = self.handler.locate_product_for_entry(session, '12345',
                                                       lookup_fields=['vendor_code'])
        self.assertIs(result, product5)

        # bad lookup field is ignored
        result = self.handler.locate_product_for_entry(session, '12345',
                                                       lookup_fields=['bogus'])
        self.assertIsNone(result)



# -*- coding: utf-8; -*-
"""add purchase.id

Revision ID: 6b5a277cc31b
Revises: 2afee42cc24d
Create Date: 2021-01-29 12:44:46.663170

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '6b5a277cc31b'
down_revision = '2afee42cc24d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase
    op.add_column('purchase', sa.Column('id', sa.Integer(), nullable=True))


def downgrade():

    # purchase
    op.drop_column('purchase', 'id')
